(*----------------------------------------------------------------------------*
 *  Direct3D sample from DirectX 9.0 SDK Summer 2003                          *
 *  Delphi adaptation by Alexey Barkovoy (e-mail: clootie@ixbt.com)           *
 *                                                                            *
 *  Delphi versions 5-7 are supported                                         *
 *                                                                            *
 *  Modified: 30-Dec-2003                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru/delphi                                         *
 *----------------------------------------------------------------------------*)
//-----------------------------------------------------------------------------
// File: EnhancedMesh.cpp
//
// Desc: Sample showing enhanced meshes in D3D
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------

{$I DirectX.inc}

unit EnhancedMeshUnit;

interface

uses
  Windows, Messages, SysUtils, CommDlg, Math, Resource,
  Direct3D9, D3DX9, D3DApp, D3DEnumeration, D3DSettings,
  D3DFile, D3DFont, D3DUtil, DXUtil;

// C++Builder compile hack
{$NOINCLUDE Resource}
{$HPPEMIT '#include "Resource.hpp"'}

type
  PD3DXMaterialsArray = ^TD3DXMaterialsArray;
  TD3DXMaterialsArray = array [0..0] of TD3DXMaterial;

  TDirect3DTexturesArray = array of IDirect3DTexture9;

  //-----------------------------------------------------------------------------
  // Name: class CMyD3DApplication
  // Desc: Application class. The base class (CD3DApplication) provides the
  //       generic functionality needed in all Direct3D samples. CMyD3DApplication
  //       adds functionality specific to this sample program.
  //-----------------------------------------------------------------------------
  CMyD3DApplication = class(CD3DApplication)
  private
    m_strMeshFilename:  array [0..511] of Char;
    m_strInitialDir:    array [0..511] of Char;

    m_pMeshSysMem:      ID3DXMesh;          // system memory version of mesh, lives through resize's
    m_pMeshEnhanced:    ID3DXMesh;          // vid mem version of mesh that is enhanced
    m_dwNumSegs:        Cardinal;           // number of segments per edge (tesselation level)
    m_pMaterials:       PD3DXMaterialsArray;// pointer to material info in m_pbufMaterials
    m_ppTextures:       TDirect3DTexturesArray;// array of textures, entries are NULL if no texture specified
    m_dwNumMaterials:   DWORD;              // number of materials

    m_bShowHelp:        Boolean;
    m_pFont:            CD3DFont;
    m_Camera:           CModelViewerCamera; // mouse rotation utility
    m_vObjectCenter:    TD3DXVector3;       // Center of bounding sphere of object
    m_fObjectRadius:    Single;             // Radius of bounding sphere of object

    m_pbufMaterials:    ID3DXBuffer;        // contains both the materials data and the filename strings
    m_pbufAdjacency:    ID3DXBuffer;        // Contains the adjacency info loaded with the mesh

    m_bUseHWNPatches:   Boolean;
    m_bWireframe:       Boolean;

    function GenerateEnhancedMesh(cNewNumSegs: Cardinal): HResult;

  protected
    function OneTimeSceneInit: HResult; override;
    function InitDeviceObjects: HResult; override;
    function ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
      adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT; override;
    function RestoreDeviceObjects: HResult; override;
    function InvalidateDeviceObjects: HResult; override;
    function DeleteDeviceObjects: HResult; override;
    function Render: HResult; overload; override;
    function FrameMove: HResult; override;
    function FinalCleanup: HResult; override;
  public
    constructor Create;
    destructor Destroy; override;
    function MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT; override;
 end;

implementation

{ CMyD3DApplication }

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
constructor CMyD3DApplication.Create;
begin
  inherited Create;

  m_strWindowTitle    := 'Enhanced Mesh - N-Patches';
  m_d3dEnumeration.AppUsesDepthBuffer := True;
  m_bShowCursorWhenFullscreen := TRUE;

  m_pMeshSysMem     := nil;
  m_pMeshEnhanced   := nil;
  m_pMaterials      := nil;
  m_ppTextures      := nil;
  m_dwNumMaterials  := 0;
  m_pbufMaterials   := nil;
  m_pbufAdjacency   := nil;
  m_dwNumSegs       := 2;
  m_bUseHWNPatches  := False;
  m_bWireframe      := False;
  m_bShowHelp       := False;

  m_pFont           := CD3DFont.Create('Arial', 9, D3DFONT_BOLD);

  DXUtil_GetDXSDKMediaPathCb(m_strInitialDir, SizeOf(m_strInitialDir));
  m_strMeshFilename := 'tiger.x';

  m_Camera          := CModelViewerCamera.Create;
end;

destructor CMyD3DApplication.Destroy;
begin
  m_Camera.Free;
  inherited;
end;

//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
function CMyD3DApplication.OneTimeSceneInit: HResult;
begin
  // Set cursor to indicate that user can move the object with the mouse
{$IFDEF WIN64}
  SetClassLongPtr(m_hWnd, GCLP_HCURSOR, LoadCursor(0, IDC_SIZEALL));
{$ELSE}
  SetClassLong(m_hWnd, GCL_HCURSOR, LoadCursor(0, IDC_SIZEALL));
{$ENDIF}
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FrameMove: HResult;
begin
  m_Camera.FrameMove(m_fElapsedTime);

  m_pd3dDevice.SetTransform(D3DTS_WORLD, m_Camera.GetWorldMatrix^);
  m_pd3dDevice.SetTransform(D3DTS_VIEW, m_Camera.GetViewMatrix^);

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.Render: HResult;
var
  fNumSegs: Single;
  i: Integer;

  Str1, Str2: String;
begin
  // Clear the backbuffer
  m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER, $00FF, 1, 0);

  // Begin the scene
  if SUCCEEDED(m_pd3dDevice.BeginScene) then
  begin
    if (m_bUseHWNPatches) then
    begin
      fNumSegs:= m_dwNumSegs;
      m_pd3dDevice.SetNPatchMode(fNumSegs);
    end;

    // set and draw each of the materials in the mesh
    for i:= 0 to (m_dwNumMaterials - 1) do
    begin
      m_pd3dDevice.SetMaterial(m_pMaterials[i].MatD3D);
      m_pd3dDevice.SetTexture(0, m_ppTextures[i]);

      m_pMeshEnhanced.DrawSubset(i);
    end;

    if m_bUseHWNPatches then
    begin
      m_pd3dDevice.SetNPatchMode(0);
    end;

    // Output stats
    begin
      // Output statistics
      m_pFont.DrawText(   2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats );
      m_pFont.DrawText(   2, 20, D3DCOLOR_ARGB(255,255,255,0), m_strDeviceStats );

      Str1:= 'NumSegs:';
      Str2:= Format('%d ', [m_dwNumSegs]);
      m_pFont.DrawText(   2, 40, $ffffff00, PChar(Str1));
      m_pFont.DrawText( 150, 40, $ffffffff, PChar(Str2));
      Str1:= 'NumFaces:';
      if not Assigned(m_pMeshEnhanced) then Str2:= #0
        else Str2:= Format('%d ', [m_pMeshEnhanced.GetNumFaces]);
      m_pFont.DrawText(   2, 60, $ffffff00, PChar(Str1));
      m_pFont.DrawText( 150, 60, $ffffffff, PChar(Str2));
      Str1:= 'NumVertices:';
      if not Assigned(m_pMeshEnhanced) then Str2:= #0
        else Str2:= Format('%d ', [m_pMeshEnhanced.GetNumVertices]);
      m_pFont.DrawText(   2, 80, $ffffff00, PChar(Str1));
      m_pFont.DrawText( 150, 80, $ffffffff, PChar(Str2));
    end;

    if m_bShowHelp then
    begin
      m_pFont.DrawText( 2, 100, D3DCOLOR_ARGB(255,255,255,255),
                                'Keyboard controls:');
      m_pFont.DrawText( 20, 120, D3DCOLOR_ARGB(255,255,255,255),
                              'Add Segments'#10'Remove Segments'#10'Wireframe'#10 +
                              'Solid'#10 +
                              'Help'#10'Change Device'#10'Quit');
      m_pFont.DrawText( 240, 120, D3DCOLOR_ARGB(255,255,255,255),
                              'Up Arrow'#10'Down Arrow'#10'W'#10 +
                              'S'#10'F1'#10'F2'#10'Esc');
    end else
    begin
      m_pFont.DrawText( 2, 100, D3DCOLOR_ARGB(255,255,255,255), 'Press F1 for help');
    end;

    // End the scene.
    m_pd3dDevice.EndScene;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: This creates all device-dependent managed objects, such as managed
//       textures and managed vertex buffers.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InitDeviceObjects: HResult;
var
  pVB: IDirect3DVertexBuffer9;
  pVertices: Pointer;
  pTempMesh: ID3DXMesh;
  strMeshPath: array[0..511] of Char;
  strTexturePath: array[0..511] of Char;
  tszName: PChar;
{$IFDEF UNICODE}
  tszBuf: array[0..MAX_PATH-1] of Char;
{$ENDIF}
  i: Integer;
begin
  pVertices := nil;

  // Initialize the font's internal textures
  m_pFont.InitDeviceObjects(m_pd3dDevice);

  // Load the mesh from the specified file
  DXUtil_FindMediaFileCb(strMeshPath, SizeOf(strMeshPath), m_strMeshFilename);

  Result:= D3DXLoadMeshFromX(strMeshPath, D3DXMESH_SYSTEMMEM, m_pd3dDevice,
                             @m_pbufAdjacency, @m_pbufMaterials, nil, @m_dwNumMaterials,
                             m_pMeshSysMem);
  if FAILED(Result) then
  begin
    // hide error so that we can display a blue screen
    Result := S_OK;
    Exit;
  end;

  // Lock the vertex buffer, to generate a simple bounding sphere
  Result:= m_pMeshSysMem.GetVertexBuffer(pVB);
  if FAILED(Result) then Exit;

  Result:= pVB.Lock(0, 0, pVertices, 0);
  if FAILED(Result) then Exit; // VB will be released by Delphi compiler

  Result:= D3DXComputeBoundingSphere(PD3DXVector3(pVertices), m_pMeshSysMem.GetNumVertices,
                                  D3DXGetFVFVertexSize(m_pMeshSysMem.GetFVF), m_vObjectCenter,
                                  m_fObjectRadius);
  if FAILED(Result) then
  begin
    pVB.Unlock;
    pVB:= nil;
    Exit;
  end;

  if (0 = m_dwNumMaterials) then
  begin
    pVB.Unlock;
    pVB:= nil;
    Result := E_INVALIDARG;
    Exit;
  end;

  // Get the array of materials out of the Result :=ed buffer, allocate a
  // texture array, and load the textures
  m_pMaterials := PD3DXMaterialsArray(m_pbufMaterials.GetBufferPointer);
  SetLength(m_ppTextures, m_dwNumMaterials);

  for i:= 0 to (m_dwNumMaterials - 1) do
  begin
{$IFDEF UNICODE}
    tszName := tszBuf;
    MultiByteToWideChar(CP_ACP, 0, m_pMaterials[i].pTextureFilename, -1, tszBuf, MAX_PATH);
    tszBuf[MAX_PATH - 1] := #0;
{$ELSE}
    tszName := m_pMaterials[i].pTextureFilename;
{$ENDIF}
    DXUtil_FindMediaFileCb(strTexturePath, SizeOf(strTexturePath), tszName);
    if FAILED(D3DXCreateTextureFromFile(m_pd3dDevice, strTexturePath, m_ppTextures[i]))
    then m_ppTextures[i]:= nil;
  end;

  pVB.Unlock;
  pVB:= nil;

  // Make sure there are normals, which are required for the tesselation
  // enhancement
  if (m_pMeshSysMem.GetFVF and D3DFVF_NORMAL) = 0 then
  begin
    Result:= m_pMeshSysMem.CloneMeshFVF(m_pMeshSysMem.GetOptions,
                                      m_pMeshSysMem.GetFVF or D3DFVF_NORMAL,
                                      m_pd3dDevice, pTempMesh);
    if FAILED(Result) then Exit;

    D3DXComputeNormals(pTempMesh, nil);

    m_pMeshSysMem:= nil;
    m_pMeshSysMem:= pTempMesh;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: GenerateEnhancedMesh()
// Desc:
//-----------------------------------------------------------------------------
function CMyD3DApplication.GenerateEnhancedMesh(cNewNumSegs: Cardinal): HResult;
var
  pMeshEnhancedSysMem: ID3DXMesh;
  pMeshTemp: ID3DXMesh;
  c: Single;
  dwMeshEnhancedFlags: DWORD;
begin
  if (m_pMeshSysMem = nil) then
  begin
    Result:= S_OK;
    Exit;
  end;

  // if using hw, just copy the mesh
  if (m_bUseHWNPatches) then
  begin
    Result:= m_pMeshSysMem.CloneMeshFVF(D3DXMESH_WRITEONLY or D3DXMESH_NPATCHES or
                (m_pMeshSysMem.GetOptions and D3DXMESH_32BIT),
                m_pMeshSysMem.GetFVF, m_pd3dDevice, pMeshTemp);
    if FAILED(Result) then Exit;
  end else  // tesselate the mesh in sw
  begin
    c:= cNewNumSegs;
    // Create an enhanced version of the mesh, will be in sysmem since source is
    Result:= D3DXTessellateNPatches(m_pMeshSysMem, PDWord(m_pbufAdjacency.GetBufferPointer),
                                    c, False, pMeshEnhancedSysMem, nil);
    if FAILED(Result) then
    begin
      // If the tessellate failed, there might have been more triangles or vertices
      // than can fit into a 16bit mesh, so try cloning to 32bit before tessellation

      Result:= m_pMeshSysMem.CloneMeshFVF(D3DXMESH_SYSTEMMEM or D3DXMESH_32BIT,
                 m_pMeshSysMem.GetFVF, m_pd3dDevice, pMeshTemp);
      if FAILED(Result) then Exit;

      Result:= D3DXTessellateNPatches(pMeshTemp, PDWord(m_pbufAdjacency.GetBufferPointer),
                 c, False, pMeshEnhancedSysMem, nil);
      if FAILED(Result) then Exit;

      pMeshTemp:= nil;
    end;

    // Make a vid mem version of the mesh
    // Only set WRITEONLY if it doesn't use 32bit indices, because those
    // often need to be emulated, which means that D3DX needs read-access.
    dwMeshEnhancedFlags := pMeshEnhancedSysMem.GetOptions and D3DXMESH_32BIT;
    if ((dwMeshEnhancedFlags and D3DXMESH_32BIT) = 0) then
      dwMeshEnhancedFlags := dwMeshEnhancedFlags or D3DXMESH_WRITEONLY;
    Result:= pMeshEnhancedSysMem.CloneMeshFVF(dwMeshEnhancedFlags, m_pMeshSysMem.GetFVF,
               m_pd3dDevice, pMeshTemp);
    if FAILED(Result) then Exit; // pMeshEnhancedSysMem:= nil - automatically

    // Latch in the enhanced mesh
    pMeshEnhancedSysMem:= nil;
  end;

  m_pMeshEnhanced := nil;
  m_pMeshEnhanced := pMeshTemp;
  m_dwNumSegs     := cNewNumSegs;

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Restore device-memory objects and state after a device is created or
//       resized.
//-----------------------------------------------------------------------------
function CMyD3DApplication.RestoreDeviceObjects: HResult;
var
  vecLightDirUnnormalized: TD3DXVector3;
  light: TD3DLight9;
  fAspect: Single;
begin
  m_bUseHWNPatches:= (m_d3dCaps.DevCaps and D3DDEVCAPS_NPATCHES) <> 0;

  Result:= GenerateEnhancedMesh(m_dwNumSegs);
  if FAILED(Result) then Exit;

  // Setup render state
  m_pd3dDevice.SetRenderState(D3DRS_LIGHTING,     iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_DITHERENABLE, iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_ZENABLE,      iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_AMBIENT,      $33333333);
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);

  // Setup the light
  vecLightDirUnnormalized := D3DXVector3(0.0, -1.0, 1.0);
  ZeroMemory(@light, SizeOf(TD3DLight9));
  light._Type       := D3DLIGHT_DIRECTIONAL;
  light.Diffuse.r   := 1.0;
  light.Diffuse.g   := 1.0;
  light.Diffuse.b   := 1.0;
  D3DXVec3Normalize(light.Direction, vecLightDirUnnormalized);
  light.Position.x   := 0.0;
  light.Position.y   := -1.0;
  light.Position.z   := 1.0;
  light.Range        := 1000.0;

  m_pd3dDevice.SetLight(0, light);
  m_pd3dDevice.LightEnable(0, True);

  // Setup the arcball parameters
  m_Camera.SetWindow(m_d3dsdBackBuffer.Width, m_d3dsdBackBuffer.Height, 0.85);

  // Setup the projection matrix
  fAspect := m_d3dsdBackBuffer.Width / m_d3dsdBackBuffer.Height;
  m_Camera.SetProjParams(D3DX_PI/4, fAspect, m_fObjectRadius/64.0, m_fObjectRadius*200.0);
  m_pd3dDevice.SetTransform(D3DTS_PROJECTION, m_Camera.GetProjMatrix^);

  // Restore the font
  m_pFont.RestoreDeviceObjects;

  if m_bWireframe
  then m_pd3dDevice.SetRenderState(D3DRS_FILLMODE, D3DFILL_WIREFRAME)
  else m_pd3dDevice.SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID);

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc: Called when the device-dependent objects are about to be lost.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InvalidateDeviceObjects: HResult;
begin
  m_pFont.InvalidateDeviceObjects;

  m_pMeshEnhanced:= nil;

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.DeleteDeviceObjects: HResult;
var
  i: Integer;
begin
  m_pFont.DeleteDeviceObjects;

  for i:= 0 to m_dwNumMaterials - 1 do m_ppTextures[i]:= nil;
  SetLength(m_ppTextures, 0);
  m_pMeshSysMem:= nil;
  m_pbufMaterials:= nil;
  SAFE_RELEASE(m_pbufAdjacency);
  m_dwNumMaterials := 0;

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FinalCleanup: HResult;
begin
  FreeAndNil(m_pFont);

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Message proc function to handle key and menu input
//-----------------------------------------------------------------------------
function CMyD3DApplication.MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM;
  lParam: LPARAM): LRESULT;
{$WRITEABLECONST ON}
const
  ofn: TOpenFilename = (
    lStructSize: SizeOf(TOpenFilename);
    lpstrFilter: '.X Files (.x)'#0'*.x'#0#0; nFilterIndex: 1;
    nMaxFile: 512;
    nMaxFileTitle: 512;
    lpstrTitle: 'Open Mesh File';
    Flags: OFN_FILEMUSTEXIST;
    nFileExtension: 1
  );
{$WRITEABLECONST OFF}
var
  g_strFilename: array [0..511] of Char;
  pLastSlash: PChar;
begin
  // Pass mouse messages to the ArcBall so it can build internal matrices
  m_Camera.HandleMessages(hWnd, uMsg, wParam, lParam);

  // Trap context menu
  if (WM_CONTEXTMENU = uMsg) then
  begin
    Result := 0;
    Exit;
  end;

  // Handle key presses
  if (WM_KEYDOWN = uMsg) then
  begin
    if (VK_UP = wParam)
    then GenerateEnhancedMesh(m_dwNumSegs + 1);

    if (VK_DOWN = wParam)
    then GenerateEnhancedMesh(Max(1, m_dwNumSegs - 1));

    if (Ord('W') = wParam) then
    begin
      m_bWireframe := True;
      m_pd3dDevice.SetRenderState(D3DRS_FILLMODE, D3DFILL_WIREFRAME);
    end;

    if (Ord('S') = wParam) then
    begin
      m_bWireframe := False;
      m_pd3dDevice.SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID);
    end;
  end
  else if (uMsg = WM_COMMAND) then
  begin
    // Handle the open file command
    case LOWORD(wParam) of
      IDM_OPENFILE:
      begin
        // Display the OpenFileName dialog. Then, try to load the specified file
        g_strFilename:= '';

        {$IFDEF COMPILER6_UP}
        // Handle Delphi6 case of extended OFN structure
        // (new members are invalid for Win98SE & below OS'es)
        ofn.lStructSize:= 76;

        {$ENDIF}
        ofn.lpstrFile:= m_strMeshFilename;
        ofn.lpstrFileTitle:= g_strFilename;
        ofn.lpstrInitialDir:= m_strInitialDir;

        if GetOpenFileName(ofn) then
        begin
          StrCopy(m_strInitialDir, m_strMeshFilename);
          pLastSlash:= StrPos(m_strInitialDir, '\');
          if (pLastSlash <> nil) then pLastSlash^ := #0;

          SetCurrentDirectory(m_strInitialDir);

          // Destroy and recreate everything
          InvalidateDeviceObjects;
          DeleteDeviceObjects;
          InitDeviceObjects;
          RestoreDeviceObjects;
        end;
      end;

      IDM_TOGGLEHELP: m_bShowHelp := not m_bShowHelp;
    end;
  end;

  // Pass remaining messages to default handler
  Result := inherited MsgProc(hWnd, uMsg, wParam, lParam);
end;

//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device initialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
function CMyD3DApplication.ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
  adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT; 
begin
  // On a non-pure device, if it can do rt-Patches,
  // it can do n-Patches as well.
  if (dwBehavior and D3DCREATE_PUREDEVICE <> 0) and
     ((pCaps.DevCaps and D3DDEVCAPS_NPATCHES = 0) and
      (pCaps.DevCaps and D3DDEVCAPS_RTPATCHES <> 0)) then
  begin
    Result:= E_FAIL;
    Exit;
  end;

  Result:= S_OK;
end;

end.

