(*----------------------------------------------------------------------------*
 *  Direct3D sample from DirectX 9.0 SDK Summer 2003                          *
 *  Delphi adaptation by Alexey Barkovoy (e-mail: clootie@ixbt.com)           *
 *                                                                            *
 *  Delphi versions 5-7 are supported                                         *
 *                                                                            *
 *  Modified: 30-Dec-2003                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru/delphi                                         *
 *----------------------------------------------------------------------------*)
//-----------------------------------------------------------------------------
// File: OptimizedMesh.cpp
//
// Desc: Sample of optimizing meshes in D3D
//
// Copyright (c) Corporation. All rights reserved.
//-----------------------------------------------------------------------------

unit OptimizedMeshUnit;

{$I DirectX.inc}

interface

uses
  Windows, Messages, SysUtils, CommDlg, Resource,
  Direct3D9, D3DX9, D3DApp, D3DEnumeration, D3DSettings,
  D3DFile, D3DFont, D3DUtil, DXUtil;

// C++Builder compile hack
{$NOINCLUDE Resource}
{$HPPEMIT '#include "Resource.hpp"'}

type
  // PSStripData = ^TSStripData;
  TSStripData = class
    m_pStrips: IDirect3DIndexBuffer9;       // strip indices (single strip)
    m_pStripsMany: IDirect3DIndexBuffer9;   // strip indices (many strips)

    m_cStripIndices: DWord;
    m_rgcStripLengths: array of DWord;
    m_cStrips: DWord;

    constructor Create;
  end;

  TSMeshData = class
    m_pMeshSysMem: ID3DXMesh;                // System memory copy of mesh

    m_pMesh: ID3DXMesh;                      // Local version of mesh, copied on resize
    m_pVertexBuffer: IDirect3DVertexBuffer9; // vertex buffer of mesh

    m_rgStripData: array of TSStripData;     // strip indices split by attribute
    m_cStripDatas: DWord;

    constructor Create;
    procedure ReleaseLocalMeshes;
    procedure ReleaseAll;
  end;

  TDirect3DTexturesArray = array [0..0] of IDirect3DTexture9;
  TD3DMaterialsArray = array [0..0] of TD3DMaterial9;

  //-----------------------------------------------------------------------------
  // Name: class CMyD3DApplication
  // Desc: Application class. The base class (CD3DApplication) provides the
  //       generic functionality needed in all Direct3D samples. CMyD3DApplication
  //       adds functionality specific to this sample program.
  //-----------------------------------------------------------------------------
  CMyD3DApplication = class(CD3DApplication)
  private
    m_strMeshFilename:  array [0..511] of Char;
    m_strInitialDir:    array [0..511] of Char;

    m_bShowVertexCacheOptimized:  BOOL;
    m_bShowStripReordered:        BOOL;
    m_bShowStrips:                BOOL;
    m_bShowSingleStrip:           BOOL;
    m_bForce32ByteFVF:            BOOL;

    m_bCantDoSingleStrip:         BOOL; // Single strip would be too many primitives

    m_pFont:            CD3DFont;               // Font for outputting frame stats

    m_ArcBall:          CD3DArcBall;            // Mouse rotation utility
    m_vObjectCenter:    TD3DXVector3;           // Center of bounding sphere of object
    m_fObjectRadius:    Single;                 // Radius of bounding sphere of object

    m_matWorld:         TD3DXMatrixA16;
    m_cObjectsPerSide:  DWord;                  // sqrt of the number of objects to draw

    m_dwMemoryOptions:  DWord;

    // various forms of mesh data
    m_MeshAttrSorted:           TSMeshData;
    m_MeshStripReordered:       TSMeshData;
    m_MeshVertexCacheOptimized: TSMeshData;

    m_dwNumMaterials:   DWord; // Number of materials

    m_pMeshTextures: array of IDirect3DTexture9;
    m_pMeshMaterials: array of TD3DMaterial9;

  public
    function OneTimeSceneInit: HResult; override;
    function InitDeviceObjects: HResult; override;
    function RestoreDeviceObjects: HResult; override;
    function InvalidateDeviceObjects: HResult; override;
    function DeleteDeviceObjects: HResult; override;
    function Render: HResult; override;
    function FrameMove: HResult; override;
    function FinalCleanup: HResult; override;

    function MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT; override;
    constructor Create;
    destructor Destroy; override;

    function LoadMeshData(out ppMesh: ID3DXMesh; out ppAdjacencyBuffer: ID3DXBuffer): HResult;
    function OptimizeMeshData(pMeshSysMem: ID3DXMesh; pAdjacencyBuffer: ID3DXBuffer; dwOptFlags: DWord; const pMeshData: TSMeshData): HResult;
    function UpdateLocalMeshes(pMeshData: TSMeshData): HResult;
    function DrawMeshData(pMeshData: TSMeshData): HResult;
  end;

implementation

type
  TD3DXMaterialsArray = array [0..0] of TD3DXMaterial;
const
  CChecked: array [False..True] of DWord = (MF_UNCHECKED, MF_CHECKED);

//-------------------------------------------------
// SStripData implementation
//-------------------------------------------------
constructor TSStripData.Create;
begin
  m_pStrips:= nil;
  m_pStripsMany:= nil;
  m_cStripIndices:= 0;
  m_rgcStripLengths:= nil;
end;


//-------------------------------------------------
// SMeshData implementation
//-------------------------------------------------
constructor TSMeshData.Create;
begin
  m_pMeshSysMem:= nil;
  m_pMesh:= nil;
  m_pVertexBuffer:= nil;
  m_rgStripData:= nil;
  m_cStripDatas:= 0;
end;

procedure TSMeshData.ReleaseLocalMeshes;
begin
  SAFE_RELEASE(m_pMesh);
  SAFE_RELEASE(m_pVertexBuffer);
end;

procedure TSMeshData.ReleaseAll;
var
  i, iStripData: Integer;
begin
  SAFE_RELEASE(m_pMeshSysMem);
  SAFE_RELEASE(m_pMesh);
  SAFE_RELEASE(m_pVertexBuffer);

  for iStripData:= 0 to (m_cStripDatas - 1) do
  begin
    SAFE_RELEASE(m_rgStripData[iStripData].m_pStrips);
    SAFE_RELEASE(m_rgStripData[iStripData].m_pStripsMany);
    // delete []m_rgStripData[iStripData].m_rgcStripLengths;
    SetLength(m_rgStripData[iStripData].m_rgcStripLengths, 0);
  end;

  // delete []m_rgStripData;
  for i:= 0 to Length(m_rgStripData) - 1 do FreeAndNil(m_rgStripData[i]);
  SetLength(m_rgStripData, 0);
  m_cStripDatas:= 0;
end;

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Constructor
//-----------------------------------------------------------------------------
constructor CMyD3DApplication.Create;
begin
  inherited Create;

  // Override base class members
  m_strWindowTitle     := 'OptimizedMesh: Optimizing Meshes in D3D';
  m_d3dEnumeration.AppUsesDepthBuffer := True;
  m_bShowCursorWhenFullscreen := True;

  // Initialize member variables
  m_bShowVertexCacheOptimized := True;
  m_bShowStripReordered := False;
  m_bShowSingleStrip := True;
  m_bShowStrips := False;
  m_bShowSingleStrip := False;
  m_bForce32ByteFVF := True;
  m_bCantDoSingleStrip := False;

  m_dwMemoryOptions := D3DXMESH_MANAGED;

  m_pFont := CD3DFont.Create('Arial', 12, D3DFONT_BOLD);

  DXUtil_GetDXSDKMediaPathCb(m_strInitialDir, SizeOf(m_strInitialDir));
  m_strMeshFilename:= 'knot.x';

  m_cObjectsPerSide := 1;

  // initialize mesh data structures
  m_dwNumMaterials := 0;
  m_pMeshTextures := nil;
  m_pMeshMaterials := nil;

  m_MeshAttrSorted :=           TSMeshData.Create;
  m_MeshStripReordered :=       TSMeshData.Create;
  m_MeshVertexCacheOptimized := TSMeshData.Create;
  m_ArcBall:= CD3DArcBall.Create;
end;

destructor CMyD3DApplication.Destroy;
begin
  FreeAndNil(m_MeshAttrSorted);
  FreeAndNil(m_MeshStripReordered);
  FreeAndNil(m_MeshVertexCacheOptimized);
  FreeAndNil(m_ArcBall);
  inherited;
end;

//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
function CMyD3DApplication.OneTimeSceneInit: HResult;
begin
  // Set cursor to indicate that user can move the object with the mouse
{$IFDEF _WIN64}
  SetClassLongPtr(m_hWnd, GCLP_HCURSOR, (LONG_PTR)LoadCursor(NULL, IDC_SIZEALL));
{$ELSE}
  SetClassLong(m_hWnd, GCL_HCURSOR, LoadCursor(0, IDC_SIZEALL));
{$ENDIF}
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FrameMove: HResult;
var
  matTemp, matView: TD3DXMatrixA16;
begin
  // Setup viewing postion from ArcBall
  D3DXMatrixTranslation(m_matWorld, -m_vObjectCenter.x,
                                    -m_vObjectCenter.y,
                                    -m_vObjectCenter.z );
  D3DXMatrixMultiply(m_matWorld, m_matWorld, m_ArcBall.GetTranslationMatrix^);
  D3DXMatrixMultiply(m_matWorld, m_matWorld, m_ArcBall.GetRotationMatrix^);

  D3DXMatrixTranslation(matTemp, -m_fObjectRadius * (m_cObjectsPerSide-1),//* 0.5f,
                                 -m_fObjectRadius * (m_cObjectsPerSide-1),//* 0.5f,
                                  0 );
  D3DXMatrixMultiply(m_matWorld, m_matWorld, matTemp);

  D3DXMatrixLookAtLH(matView, D3DXVector3(0, 0,-3.7 * m_fObjectRadius * m_cObjectsPerSide),
                              D3DXVector3Zero,
                              D3DXVector3(0, 1, 0));
  m_pd3dDevice.SetTransform(D3DTS_VIEW,  matView);

  Result:= S_OK;
end;

function CMyD3DApplication.DrawMeshData(pMeshData: TSMeshData): HResult;
var
  iCurFace: DWORD;
  iMaterial: Integer;

  dwFVF:                DWORD;
  cBytesPerVertex:      DWORD;
  iStrip:               DWORD;
begin
  // Set and draw each of the materials in the mesh
  for iMaterial:= 0 to (m_dwNumMaterials - 1) do
  begin
    m_pd3dDevice.SetMaterial(m_pMeshMaterials[iMaterial]);
    m_pd3dDevice.SetTexture(0, m_pMeshTextures[iMaterial]);

    // if( !m_bShowStrips && !m_bShowSingleStrip)
    if (not m_bShowStrips and not m_bShowSingleStrip) then
    begin
      pMeshData.m_pMesh.DrawSubset(iMaterial);
    end else  // drawing strips
    begin
      dwFVF := pMeshData.m_pMesh.GetFVF;
      cBytesPerVertex := D3DXGetFVFVertexSize(dwFVF);

      m_pd3dDevice.SetFVF(dwFVF);
      m_pd3dDevice.SetStreamSource(0, pMeshData.m_pVertexBuffer, 0, cBytesPerVertex);

      if m_bShowSingleStrip then
      begin
        if not m_bCantDoSingleStrip then
        begin
          m_pd3dDevice.SetIndices(pMeshData.m_rgStripData[iMaterial].m_pStrips);

          Result:= m_pd3dDevice.DrawIndexedPrimitive(D3DPT_TRIANGLESTRIP, 0,
                                  0, pMeshData.m_pMesh.GetNumVertices,
                                  0, pMeshData.m_rgStripData[iMaterial].m_cStripIndices - 2);
          if FAILED(Result) then Exit;
        end;
      end else
      begin
        m_pd3dDevice.SetIndices(pMeshData.m_rgStripData[iMaterial].m_pStripsMany);

        iCurFace := 0;
        for iStrip:= 0 to (pMeshData.m_rgStripData[iMaterial].m_cStrips - 1) do
        begin
          Result:= m_pd3dDevice.DrawIndexedPrimitive(D3DPT_TRIANGLESTRIP, 0,
                                  0, pMeshData.m_pMesh.GetNumVertices,
                                  iCurFace, pMeshData.m_rgStripData[iMaterial].m_rgcStripLengths[iStrip]);
          if FAILED(Result) then Exit;

          iCurFace:= iCurFace + 2 + pMeshData.m_rgStripData[iMaterial].m_rgcStripLengths[iStrip];
        end;
      end;
    end;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.Render: HResult;
var
  xOffset: DWord;
  yOffset: DWord;
  matWorld, matTemp: TD3DXmatrixA16;
  cTriangles: DWord;
  fTrisPerSec: Single;
  strInfo: array [0..119] of Char;
  szOptString: PChar;
begin
  // cTriangles := 0; //BAA: not used anyway

  // Clear the scene
  m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER,
                     $000000ff, 1.0, $00000000);

  // Draw scene
  if SUCCEEDED(m_pd3dDevice.BeginScene) then
  begin
    for xOffset:= 0 to m_cObjectsPerSide - 1 do
    begin
      for yOffset:= 0 to m_cObjectsPerSide - 1 do
      begin
        D3DXMatrixTranslation(matTemp, m_fObjectRadius * xOffset * 2,
                                       m_fObjectRadius * yOffset * 2,
                                       0);
        D3DXMatrixMultiply(matWorld, m_matWorld, matTemp);
        m_pd3dDevice.SetTransform(D3DTS_WORLD, matWorld);

        if m_bShowVertexCacheOptimized
        then
          DrawMeshData(m_MeshVertexCacheOptimized)
        else
          if m_bShowStripReordered
          then DrawMeshData(m_MeshStripReordered)
          else DrawMeshData(m_MeshAttrSorted);
      end;
    end;

    // Calculate and show triangles per sec, a reasonable throughput number
    if (m_MeshAttrSorted.m_pMesh <> nil) then
      cTriangles := m_MeshAttrSorted.m_pMesh.GetNumFaces * m_cObjectsPerSide * m_cObjectsPerSide
    else
      cTriangles := 0;

    fTrisPerSec := m_fFPS * cTriangles;

    if m_bShowVertexCacheOptimized then szOptString:= 'VCache Optimized'
    else if m_bShowStripReordered then  szOptString:= 'Strip Reordered'
    else szOptString:= 'Unoptimized';

    // Output statistics
    StrFmt(strInfo, '%s, %d tris per sec, %d triangles',
      [szOptString, Trunc(fTrisPerSec), cTriangles]);

    m_pFont.DrawText(2, 0,  D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats);
    m_pFont.DrawText(2, 20, D3DCOLOR_ARGB(255,255,255,0), m_strDeviceStats);
    m_pFont.DrawText(2, 40, D3DCOLOR_ARGB(255,255,255,0), strInfo);
    if (m_bShowSingleStrip and m_bCantDoSingleStrip) then
    begin
      StrCopy(strInfo, 'Couldn''t draw to single strip -- too many primitives');
      m_pFont.DrawText(2, 60, D3DCOLOR_ARGB(255,255,0,0), strInfo);
    end;

    m_pd3dDevice.EndScene;
  end;

  Result:= S_OK;
end;


function CMyD3DApplication.LoadMeshData(out ppMesh: ID3DXMesh;
  out ppAdjacencyBuffer: ID3DXBuffer): HResult;
var
  pMeshVB:              IDirect3DVertexBuffer9;
  pD3DXMtrlBuffer:      ID3DXBuffer;
  pVertices:            Pointer;
  strMesh:              array[0..511] of Char;
  bNormalsInFile:       BOOL;
  pMeshSysMem:          ID3DXMesh;
  pMeshTemp:            ID3DXMesh;
  i:                    Integer;
  d3dxMaterials:        ^TD3DXMaterialsArray;
  strPath:              array[0..511] of Char;
  tszFilename:          PChar;
{$IFDEF UNICODE}
  tszBuf: array[0..MAX_PATH-1] of Char;
{$ENDIF}
begin
  pMeshVB:= nil;
  pD3DXMtrlBuffer:= nil;
  // Result:= S_OK;
  pMeshSysMem:= nil;

  // Get a path to the media file
  Result:= DXUtil_FindMediaFileCb(strMesh, SizeOf(strMesh), m_strMeshFilename);
  if FAILED(Result) then Exit;

  // Load the mesh from the specified file
  Result:= D3DXLoadMeshFromX(strMesh, D3DXMESH_SYSTEMMEM, m_pd3dDevice,
                             @ppAdjacencyBuffer, @pD3DXMtrlBuffer, nil,
                             @m_dwNumMaterials, pMeshSysMem);
  if FAILED(Result) then Exit;

  // Get the array of materials out of the returned buffer, and allocate a texture array
  d3dxMaterials := pD3DXMtrlBuffer.GetBufferPointer;
  try
    SetLength(m_pMeshMaterials, m_dwNumMaterials);
    SetLength(m_pMeshTextures,  m_dwNumMaterials);
  except
    Result:= E_OUTOFMEMORY;
    Exit;
  end;

  for i:= 0 to m_dwNumMaterials - 1 do
  begin
    m_pMeshMaterials[i] := d3dxMaterials[i].MatD3D;
    m_pMeshMaterials[i].Ambient := m_pMeshMaterials[i].Diffuse;
    m_pMeshTextures[i]  := nil;

    // Get a path to the texture
    if (d3dxMaterials[i].pTextureFilename <> nil) then
    begin
{$IFDEF UNICODE}
      tszFilename := tszBuf;
      MultiByteToWideChar(CP_ACP, 0, d3dxMaterials[i].pTextureFilename, -1, @tszBuf, MAX_PATH);
      tszBuf[MAX_PATH - 1] := #0;
{$ELSE}
      tszFilename := d3dxMaterials[i].pTextureFilename;
{$ENDIF}
      DXUtil_FindMediaFileCb(strPath, SizeOf(strPath), tszFilename);

      // Load the texture
      D3DXCreateTextureFromFile(m_pd3dDevice, strPath, m_pMeshTextures[i]);
    end;
  end;

  // Done with the material buffer
  SAFE_RELEASE(pD3DXMtrlBuffer);

  // Lock the vertex buffer, to generate a simple bounding sphere
  Result:= pMeshSysMem.GetVertexBuffer(pMeshVB);
  if SUCCEEDED(Result) then
  begin
    Result:= pMeshVB.Lock(0, 0, pVertices, D3DLOCK_NOSYSLOCK);
    if SUCCEEDED(Result) then
    begin
      Result:= D3DXComputeBoundingSphere(PD3DXVector3(pVertices), pMeshSysMem.GetNumVertices,
                                         D3DXGetFVFVertexSize(pMeshSysMem.GetFVF),
                                         m_vObjectCenter, m_fObjectRadius);
      pMeshVB.Unlock;
    end;
    pMeshVB:= nil;
  end;
  if FAILED(Result) then Exit;

  // remember if there were normals in the file, before possible clone operation
  bNormalsInFile := (pMeshSysMem.GetFVF and D3DFVF_NORMAL) <> 0;

  // if using 32byte vertices, check fvf
  if m_bForce32ByteFVF then
  begin
    // force 32 byte vertices
    if (pMeshSysMem.GetFVF <> (D3DFVF_XYZ or D3DFVF_NORMAL or D3DFVF_TEX1)) then
    begin
      Result:= pMeshSysMem.CloneMeshFVF(pMeshSysMem.GetOptions, D3DFVF_XYZ or D3DFVF_NORMAL or D3DFVF_TEX1,
                                        m_pd3dDevice, pMeshTemp);
      if FAILED(Result) then Exit;

      pMeshSysMem := pMeshTemp;
    end;
  end
  // otherwise, just make sure that there is a normal mesh
  else if ((pMeshSysMem.GetFVF and D3DFVF_NORMAL) = 0) then
  begin
    Result:= pMeshSysMem.CloneMeshFVF(pMeshSysMem.GetOptions, pMeshSysMem.GetFVF or D3DFVF_NORMAL,
                                      m_pd3dDevice, pMeshTemp);
    if FAILED(Result) then Exit;

    pMeshSysMem := pMeshTemp;
  end;

  // Compute normals for the mesh, if not present
  if not bNormalsInFile then D3DXComputeNormals(pMeshSysMem, nil);

  ppMesh:= pMeshSysMem;
  pMeshSysMem:= nil;
end;

function CMyD3DApplication.OptimizeMeshData(pMeshSysMem: ID3DXMesh;
  pAdjacencyBuffer: ID3DXBuffer; dwOptFlags: DWord; const pMeshData: TSMeshData): HResult;
var
  pbufTemp: ID3DXBuffer;
  iMaterial, i: DWord;
  primCount: LongWord;
begin
  // attribute sort - the un-optimized mesh option
  //          remember the adjacency for the vertex cache optimization
  Result:= pMeshSysMem.Optimize(dwOptFlags or D3DXMESH_SYSTEMMEM,
                                pAdjacencyBuffer.GetBufferPointer,
                                nil, nil, nil, pMeshData.m_pMeshSysMem);
  if FAILED(Result) then Exit;

  pMeshData.m_cStripDatas := m_dwNumMaterials;
  try
    SetLength(pMeshData.m_rgStripData, pMeshData.m_cStripDatas);
  except
    Result:= E_OUTOFMEMORY;
    Exit;
  end;

  for i:= 0 to pMeshData.m_cStripDatas - 1 do
    pMeshData.m_rgStripData[i]:= TSStripData.Create;

  m_bCantDoSingleStrip := False;
  for iMaterial:= 0 to (m_dwNumMaterials - 1) do
  begin
    Result:= D3DXConvertMeshSubsetToSingleStrip(
               pMeshData.m_pMeshSysMem, iMaterial, D3DXMESH_IB_MANAGED,
               pMeshData.m_rgStripData[iMaterial].m_pStrips,
               @pMeshData.m_rgStripData[iMaterial].m_cStripIndices);
    if FAILED(Result) then Exit;

    primCount := pMeshData.m_rgStripData[iMaterial].m_cStripIndices - 2;
    if (primCount > m_d3dCaps.MaxPrimitiveCount) then m_bCantDoSingleStrip := True;

    Result:= D3DXConvertMeshSubsetToStrips(
               pMeshData.m_pMeshSysMem, iMaterial, D3DXMESH_IB_MANAGED,
               pMeshData.m_rgStripData[iMaterial].m_pStripsMany, nil,
               @pbufTemp, @pMeshData.m_rgStripData[iMaterial].m_cStrips);
    if FAILED(Result) then Exit;

    try
      SetLength(pMeshData.m_rgStripData[iMaterial].m_rgcStripLengths, pMeshData.m_rgStripData[iMaterial].m_cStrips);
    except
      Result:= E_OUTOFMEMORY;
      Exit;
    end;

    Move(pbufTemp.GetBufferPointer^, pMeshData.m_rgStripData[iMaterial].m_rgcStripLengths[0],
      SizeOf(DWORD)*pMeshData.m_rgStripData[iMaterial].m_cStrips);
  end;
end;


//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InitDeviceObjects: HResult;
var
  pMeshSysMem: ID3DXMesh;
  pAdjacencyBuffer: ID3DXBuffer;
begin
  // pMeshSysMem := nil;
  // pAdjacencyBuffer := nil;

  // Initialize the font
  m_pFont.InitDeviceObjects(m_pd3dDevice);

  // check current display setting
  CheckMenuItem(GetMenu(m_hWnd), ID_OPTIONS_DISPLAY1 + (m_cObjectsPerSide-1), MF_CHECKED);
  CheckMenuItem(GetMenu(m_hWnd), IDM_SHOWNONOPTIMIZEDMESH, CChecked[not m_bShowStripReordered and not m_bShowVertexCacheOptimized]);
  CheckMenuItem(GetMenu(m_hWnd), IDM_SHOWVCACHEOPTIMIZED,  CChecked[m_bShowVertexCacheOptimized]);
  CheckMenuItem(GetMenu(m_hWnd), IDM_SHOWSTRIPREORDERED,   CChecked[m_bShowStripReordered]);
  CheckMenuItem(GetMenu(m_hWnd), IDM_SHOWTRILIST,          CChecked[not m_bShowStrips and not m_bShowSingleStrip]);
  CheckMenuItem(GetMenu(m_hWnd), IDM_SHOWONESTRIP,         CChecked[m_bShowSingleStrip]);
  CheckMenuItem(GetMenu(m_hWnd), IDM_SHOWMANYSTRIPS,       CChecked[m_bShowStrips]);

  CheckMenuItem(GetMenu(m_hWnd), IDM_DYNAMICVB,            CChecked[m_dwMemoryOptions = D3DXMESH_DYNAMIC]);
  CheckMenuItem(GetMenu(m_hWnd), IDM_FORCE32BYTEVERTEX,    CChecked[m_bForce32ByteFVF]);

  Result:= LoadMeshData(pMeshSysMem, pAdjacencyBuffer);
  if FAILED(Result) then
  begin // ignore load errors, just draw blank screen if mesh is invalid
    Result:= S_OK;
    Exit;
  end;

  Result:= OptimizeMeshData(pMeshSysMem, pAdjacencyBuffer, D3DXMESHOPT_ATTRSORT, m_MeshAttrSorted);
  if FAILED(Result) then Exit;

  Result:= OptimizeMeshData(pMeshSysMem, pAdjacencyBuffer, D3DXMESHOPT_STRIPREORDER, m_MeshStripReordered);
  if FAILED(Result) then Exit;

  Result:= OptimizeMeshData(pMeshSysMem, pAdjacencyBuffer, D3DXMESHOPT_VERTEXCACHE, m_MeshVertexCacheOptimized);
end;

function CMyD3DApplication.UpdateLocalMeshes(pMeshData: TSMeshData): HResult;
begin
  Result:= S_OK;

  // if a mesh was loaded, update the local meshes
  if (pMeshData.m_pMeshSysMem <> nil) then
  begin
    Result:= pMeshData.m_pMeshSysMem.CloneMeshFVF(m_dwMemoryOptions or D3DXMESH_VB_WRITEONLY,
                                       pMeshData.m_pMeshSysMem.GetFVF,
                                       m_pd3dDevice, pMeshData.m_pMesh);
    if FAILED(Result) then Exit;

    Result:= pMeshData.m_pMesh.GetVertexBuffer(pMeshData.m_pVertexBuffer);
    if FAILED(Result) then Exit;
  end;
end;


//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.RestoreDeviceObjects: HResult;
 function D3DColorValue(r, g, b: Single; a: Single = 1): TD3DColorValue;
 begin
   Result.r:= r;
   Result.g:= g;
   Result.b:= b;
   Result.a:= a;
 end;
var
  light: TD3DLight9;
  fAspect: Single;
  matProj: TD3DXMatrix;
const
  FLT_MAX: Single       = 3.402823466e+38;  // max single value
begin
  m_pFont.RestoreDeviceObjects;

  // Setup render state
  m_pd3dDevice.SetRenderState(D3DRS_LIGHTING,     Integer(TRUE));
  m_pd3dDevice.SetRenderState(D3DRS_DITHERENABLE, Integer(TRUE));
  m_pd3dDevice.SetRenderState(D3DRS_ZENABLE,      Integer(TRUE));
  m_pd3dDevice.SetSamplerState( 0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
  m_pd3dDevice.SetSamplerState( 0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);

  // Setup the light
  light._Type        := D3DLIGHT_DIRECTIONAL;
  light.Diffuse      := D3DColorValue(1, 1, 1, 0);
  light.Specular     := D3DColorValue(0, 0, 0, 0);
  light.Ambient      := D3DColorValue(0.3, 0.3, 0.3, 0);
  light.Position     := D3DXVector3(0.0, 0.0, 0.0);
  D3DXVec3Normalize(light.Direction, D3DXVector3(0.3, -1.0, 1.0));
  light.Attenuation0 := 0.0;
  light.Attenuation1 := 0.0;
  light.Attenuation2 := 0.0;
  light.Range        := Sqrt(FLT_MAX);
  m_pd3dDevice.SetLight(0, light);
  m_pd3dDevice.LightEnable(0, TRUE);

  m_ArcBall.SetWindow(m_d3dsdBackBuffer.Width, m_d3dsdBackBuffer.Height, 0.85);
  m_ArcBall.SetTranslationRadius(m_fObjectRadius);

  fAspect:= m_d3dsdBackBuffer.Width / m_d3dsdBackBuffer.Height;

  D3DXMatrixPerspectiveFovLH(matProj, D3DX_PI/4, fAspect, m_fObjectRadius/64.0,
                             m_fObjectRadius*200.0);
  m_pd3dDevice.SetTransform(D3DTS_PROJECTION, matProj);

  // update the local copies of the meshes
  UpdateLocalMeshes(m_MeshAttrSorted);
  UpdateLocalMeshes(m_MeshStripReordered);
  UpdateLocalMeshes(m_MeshVertexCacheOptimized);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects
// Desc:
//-----------------------------------------------------------------------------
function CMyD3DApplication.InvalidateDeviceObjects: HResult;
begin
  m_pFont.InvalidateDeviceObjects;

  // update the local copies of the meshes
  m_MeshAttrSorted.ReleaseLocalMeshes;
  m_MeshStripReordered.ReleaseLocalMeshes;
  m_MeshVertexCacheOptimized.ReleaseLocalMeshes;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.DeleteDeviceObjects: HResult;
var
  i: Integer;
begin
  m_pFont.DeleteDeviceObjects;

  for i:= 0 to m_dwNumMaterials - 1 do
    m_pMeshTextures[i]:= nil;

  m_pMeshMaterials:= nil;
  m_pMeshTextures:= nil;

  m_MeshAttrSorted.ReleaseAll;
  m_MeshStripReordered.ReleaseAll;
  m_MeshVertexCacheOptimized.ReleaseAll;

  m_dwNumMaterials:= 0;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FinalCleanup
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FinalCleanup: HResult;
begin
  FreeAndNil(m_pFont);
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: MsgProc
// Desc: Message proc function to handle key and menu input
//-----------------------------------------------------------------------------
function CMyD3DApplication.MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM;
  lParam: LPARAM): LRESULT;
var
  g_strFilename: array [0..511] of Char;
  pLastSlash: PChar;
{$WRITEABLECONST ON}
const
  ofn: TOpenFilename = (
    lStructSize: SizeOf(TOpenFilename);
    hWndOwner: 0;
    hInstance: 0;
    lpstrFilter: '.X Files (.x)'#0'*.x'#0#0;
    lpstrCustomFilter: nil;
    nMaxCustFilter: 0;
    nFilterIndex: 1;
    lpstrFile: nil; // updated by code later
    nMaxFile: 512;
    lpstrFileTitle: nil; // updated by code later
    nMaxFileTitle: 512;
    lpstrInitialDir: nil; // updated by code later
    lpstrTitle: 'Open Mesh File';
    Flags: OFN_FILEMUSTEXIST;
    nFileOffset: 0;
    nFileExtension: 1;
    lpstrDefExt: nil;
    lCustData: 0;
    lpfnHook: nil;
    lpTemplateName: nil
  );
{$WRITEABLECONST OFF}
begin
  // Pass mouse messages to the ArcBall so it can build internal matrices
  m_ArcBall.HandleMessages(hWnd, uMsg, wParam, lParam);

  // Trap the context menu
  if (WM_CONTEXTMENU = uMsg) then
  begin
    Result:= 0;
    Exit;
  end;

  if (uMsg = WM_COMMAND) then
  begin
    // Toggle mesh optimization
    if (LongRec(wParam).Lo = IDM_SHOWNONOPTIMIZEDMESH) then
    begin
      m_bShowVertexCacheOptimized := FALSE;
      m_bShowStripReordered := FALSE;

      CheckMenuItem(GetMenu(hWnd), IDM_SHOWNONOPTIMIZEDMESH, MF_CHECKED);
      CheckMenuItem(GetMenu(hWnd), IDM_SHOWVCACHEOPTIMIZED, MF_UNCHECKED);
      CheckMenuItem(GetMenu(hWnd), IDM_SHOWSTRIPREORDERED, MF_UNCHECKED);
    end
    else if (LOWORD(wParam) = IDM_SHOWVCACHEOPTIMIZED) then
    begin
      m_bShowVertexCacheOptimized := TRUE;
      m_bShowStripReordered := FALSE;

      CheckMenuItem(GetMenu(hWnd), IDM_SHOWNONOPTIMIZEDMESH, MF_UNCHECKED);
      CheckMenuItem(GetMenu(hWnd), IDM_SHOWVCACHEOPTIMIZED, MF_CHECKED);
      CheckMenuItem(GetMenu(hWnd), IDM_SHOWSTRIPREORDERED, MF_UNCHECKED);
    end
    else if (LOWORD(wParam) = IDM_SHOWSTRIPREORDERED) then
    begin
      m_bShowVertexCacheOptimized := FALSE;
      m_bShowStripReordered := TRUE;

      CheckMenuItem(GetMenu(hWnd), IDM_SHOWNONOPTIMIZEDMESH, MF_UNCHECKED);
      CheckMenuItem(GetMenu(hWnd), IDM_SHOWVCACHEOPTIMIZED, MF_UNCHECKED);
      CheckMenuItem(GetMenu(hWnd), IDM_SHOWSTRIPREORDERED, MF_CHECKED);
    end
    // Toggle strips
    else if (LOWORD(wParam) = IDM_SHOWTRILIST) then
    begin
      m_bShowStrips := FALSE;
      m_bShowSingleStrip := FALSE;

      CheckMenuItem(GetMenu(hWnd), IDM_SHOWTRILIST, MF_CHECKED);
      CheckMenuItem(GetMenu(hWnd), IDM_SHOWONESTRIP, MF_UNCHECKED);
      CheckMenuItem(GetMenu(hWnd), IDM_SHOWMANYSTRIPS, MF_UNCHECKED);
    end
    else if (LOWORD(wParam) = IDM_SHOWONESTRIP) then
    begin
      m_bShowStrips := FALSE;
      m_bShowSingleStrip := TRUE;

      CheckMenuItem(GetMenu(hWnd), IDM_SHOWTRILIST, MF_UNCHECKED);
      CheckMenuItem(GetMenu(hWnd), IDM_SHOWONESTRIP, MF_CHECKED);
      CheckMenuItem(GetMenu(hWnd), IDM_SHOWMANYSTRIPS, MF_UNCHECKED);
    end
    else if (LOWORD(wParam) = IDM_SHOWMANYSTRIPS) then
    begin
      m_bShowStrips := TRUE;
      m_bShowSingleStrip := FALSE;

      CheckMenuItem(GetMenu(hWnd), IDM_SHOWTRILIST, MF_UNCHECKED);
      CheckMenuItem(GetMenu(hWnd), IDM_SHOWONESTRIP, MF_UNCHECKED);
      CheckMenuItem(GetMenu(hWnd), IDM_SHOWMANYSTRIPS, MF_CHECKED);
    end
    // Toggle vertex buffer mode
    else if (LOWORD(wParam) = IDM_DYNAMICVB) then
    begin
      if (m_dwMemoryOptions = D3DXMESH_DYNAMIC) then
      begin
        m_dwMemoryOptions := D3DXMESH_MANAGED;
        CheckMenuItem(GetMenu(hWnd), IDM_DYNAMICVB, MF_UNCHECKED);
      end else
      begin
        m_dwMemoryOptions := D3DXMESH_DYNAMIC;
        CheckMenuItem(GetMenu(hWnd), IDM_DYNAMICVB, MF_CHECKED);
      end;
      // Destroy and recreate everything
      InvalidateDeviceObjects;
      RestoreDeviceObjects;
    end
    else if (LOWORD(wParam) = IDM_FORCE32BYTEVERTEX) then
    begin
      m_bForce32ByteFVF := not m_bForce32ByteFVF;

      CheckMenuItem(GetMenu(hWnd), IDM_FORCE32BYTEVERTEX, CChecked[m_bForce32ByteFVF]);

      // Destroy and recreate everything
      InvalidateDeviceObjects;
      DeleteDeviceObjects;
      InitDeviceObjects;
      RestoreDeviceObjects;
    end

    // Handle the open file command
    else if (LongRec(wParam).Lo = IDM_OPENFILE) then
    begin
      // Display the OpenFileName dialog. Then, try to load the specified file
      g_strFilename:= '';

      {$IFDEF COMPILER6_UP}
      // Handle Delphi6 case of extended OFN structure
      // (new members are invalid for Win98SE & below OS'es)
      ofn.lStructSize:= 76;

      {$ENDIF}
      ofn.lpstrFile:= m_strMeshFilename;
      ofn.lpstrFileTitle:= g_strFilename;
      ofn.lpstrInitialDir:= m_strInitialDir;

      if (GetOpenFileName(ofn)) then
      begin
        StrCopy(m_strInitialDir, m_strMeshFilename);
        pLastSlash:= StrPos(m_strInitialDir, '\');
        if Assigned(pLastSlash) then pLastSlash[0]:= #0;
        SetCurrentDirectory(m_strInitialDir);

        // Destroy and recreate everything
        InvalidateDeviceObjects;
        DeleteDeviceObjects;
        InitDeviceObjects;
        RestoreDeviceObjects;
      end;
    end

    else if (LongRec(wParam).Lo >= ID_OPTIONS_DISPLAY1) and
            (LongRec(wParam).Lo <= ID_OPTIONS_DISPLAY36) then
    begin
      // uncheck old item
      CheckMenuItem(GetMenu(hWnd), ID_OPTIONS_DISPLAY1 + (m_cObjectsPerSide-1), MF_UNCHECKED);

      // calc new item
      m_cObjectsPerSide:= LongRec(wParam).Lo - ID_OPTIONS_DISPLAY1 + 1;

      // check new item
      CheckMenuItem(GetMenu(hWnd), ID_OPTIONS_DISPLAY1 + (m_cObjectsPerSide-1), MF_CHECKED);
    end;
  end;

  Result:= inherited MsgProc(hWnd, uMsg, wParam, lParam);
end;

end.

