(*----------------------------------------------------------------------------*
 *  Direct3D sample from DirectX 9.0 SDK Summer 2003                          *
 *  Delphi adaptation by Alexey Barkovoy (e-mail: clootie@ixbt.com)           *
 *                                                                            *
 *  Delphi versions 5-7 are supported                                         *
 *                                                                            *
 *  Modified: 30-Dec-2003                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru/delphi                                         *
 *----------------------------------------------------------------------------*)
//-----------------------------------------------------------------------------
// File: ProgressiveMesh.cpp
//
// Desc: Sample of creating progressive meshes in D3D
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------

unit ProgressiveMeshUnit;

{$I DirectX.inc}

interface

uses
  Windows, Messages, SysUtils, CommDlg, Math, Resource,
  Direct3D9, D3DX9, D3DApp, D3DEnumeration, D3DSettings,
  D3DFile, D3DFont, D3DUtil, DXUtil;

type
  TID3DXPMeshArray = array [0..0] of ID3DXPMesh;
  TDirect3DTexturesArray = array [0..0] of IDirect3DTexture9;
  TD3DMaterialsArray = array [0..0] of TD3DMaterial9;

  //-----------------------------------------------------------------------------
  // Name: class CMyD3DApplication
  // Desc: Main class to run this application. Most functionality is inherited
  //       from the CD3DApplication base class.
  //-----------------------------------------------------------------------------
  CMyD3DApplication = class(CD3DApplication)
  private
    m_strInitialDir:    array [0..511] of Char;
    m_strMeshFilename:  array [0..511] of Char; // Filename of mesh

    m_pPMeshes:         ^TID3DXPMeshArray;
    m_pPMeshFull:       ID3DXPMesh;
    m_cPMeshes:         DWORD;
    m_iPMeshCur:        DWORD;

    m_mtrlMeshMaterials:^TD3DMaterialsArray;
    m_pMeshTextures:    ^TDirect3DTexturesArray;// Array of textures, entries are NULL if no texture specified
    m_dwNumMaterials:   DWORD;                  // Number of materials

    m_Camera:           CModelViewerCamera;     // Mouse rotation utility
    m_vObjectCenter:    TD3DXVector3;           // Center of bounding sphere of object
    m_fObjectRadius:    Single;                 // Radius of bounding sphere of object

    m_pFont:            CD3DFont;               // Font for displaying help
    m_bDisplayHelp:     Boolean;
    m_bShowOptimized:   Boolean;

    m_bInitDone:        Boolean;                // hold off on any reactino to messages until fully inited
  public
    function OneTimeSceneInit: HResult; override;
    function InitDeviceObjects: HResult; override;
    function RestoreDeviceObjects: HResult; override;
    function InvalidateDeviceObjects: HResult; override;
    function DeleteDeviceObjects: HResult; override;
    function Render: HResult; override;
    function FrameMove: HResult; override;
    function FinalCleanup: HResult; override;

    function MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM; lParam: LPARAM ): LRESULT; override;
    procedure SetNumVertices(dwNumVertices: DWORD);

    constructor Create;
    destructor Destroy; override;
  end;

implementation

const
  IsChecked: array[False..True] of Integer = (MF_UNCHECKED, MF_CHECKED);

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Constructor
//-----------------------------------------------------------------------------
constructor CMyD3DApplication.Create;
begin
  inherited Create;
  
  // Override base class members
  m_strWindowTitle    := 'ProgressiveMesh: Using Progressive Meshes in D3D';
  m_d3dEnumeration.AppUsesDepthBuffer := True;
  m_bShowCursorWhenFullscreen:= True;

  // Initialize member variables
  DXUtil_GetDXSDKMediaPathCb(m_strInitialDir, SizeOf(m_strInitialDir));
  m_strMeshFilename := 'Tiger.x';

  m_pPMeshes           := nil;
  m_cPMeshes           := 0;
  m_pPMeshFull         := nil;
  m_bShowOptimized     := True;

  m_dwNumMaterials     := 0;
  m_mtrlMeshMaterials  := nil;
  m_pMeshTextures      := nil;

  m_pFont              := CD3DFont.Create('Arial', 12, D3DFONT_BOLD);
  m_bDisplayHelp       := False;

  m_bInitDone          := False;

  m_Camera            := CModelViewerCamera.Create;
end;

destructor CMyD3DApplication.Destroy;
begin
  FreeAndNil(m_Camera);
  inherited;
end;

//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
function CMyD3DApplication.OneTimeSceneInit: HResult;
begin
  // Set cursor to indicate that user can move the object with the mouse
{$IFDEF _WIN64}
  SetClassLongPtr(m_hWnd, GCLP_HCURSOR, (LONG_PTR)LoadCursor(NULL, IDC_SIZEALL));
{$ELSE}
  SetClassLong(m_hWnd, GCL_HCURSOR, LoadCursor(0, IDC_SIZEALL));
{$ENDIF}
  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FrameMove: HResult;
begin
  m_Camera.FrameMove(m_fElapsedTime);

  m_pd3dDevice.SetTransform(D3DTS_WORLD, m_Camera.GetWorldMatrix^);
  m_pd3dDevice.SetTransform(D3DTS_VIEW, m_Camera.GetViewMatrix^);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: Render
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.Render: HResult;
var
  i: Integer;
  strBuffer: array [0..511] of Char;
begin
  Result:= S_OK;
  if (not m_bInitDone) then Exit;

  // Clear the backbuffer
  m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER,
                       $000000ff, 1.0, $00000000);

  // Begin scene rendering
  if SUCCEEDED(m_pd3dDevice.BeginScene) then
  begin
    if (m_pPMeshes <> nil) then
    begin
      // Set and draw each of the materials in the mesh
      for i:= 0 to m_dwNumMaterials - 1 do
      begin
        m_pd3dDevice.SetMaterial(m_mtrlMeshMaterials[i]);
        m_pd3dDevice.SetTexture(0, m_pMeshTextures[i]);
        if m_bShowOptimized
          then m_pPMeshes[m_iPMeshCur].DrawSubset(i)
          else m_pPMeshFull.DrawSubset(i);
      end;
    end;


    // Output statistics
    m_pFont.DrawText( 2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats);
    m_pFont.DrawText( 2, 20, D3DCOLOR_ARGB(255,255,255,0), m_strDeviceStats);
    if m_bShowOptimized then
    begin
      if (m_pPMeshes = nil) then
      begin
        StrFmt(strBuffer, 'Unoptimized', []);
        m_pFont.DrawText( 2, 60, $ffffff00, strBuffer);

        StrFmt(strBuffer, 'Num Vertices = %d, Optimized %f', [0, 0.0]);
        m_pFont.DrawText( 2, 40, $ffffff00, strBuffer);
      end else
      begin
        StrFmt(strBuffer, 'PMesh %d out of %d', [m_iPMeshCur + 1, m_cPMeshes]);
        m_pFont.DrawText( 2, 60, $ffffff00, strBuffer);

        StrFmt(strBuffer, 'Num Vertices = %d, Min = %d, Max = %d%', [
               m_pPMeshes[m_iPMeshCur].GetNumVertices,
               m_pPMeshes[m_iPMeshCur].GetMinVertices,
               m_pPMeshes[m_iPMeshCur].GetMaxVertices]);
        m_pFont.DrawText( 2, 40, $ffffff00, strBuffer);
      end;
    end else
    begin
      StrFmt(strBuffer, 'Unoptimized', []);
      m_pFont.DrawText( 2, 60, $ffffff00, strBuffer);

      StrFmt(strBuffer, 'Num Vertices = %d, Min = %d, Max = %d%', [
             m_pPMeshFull.GetNumVertices,
             m_pPMeshFull.GetMinVertices,
             m_pPMeshFull.GetMaxVertices]);
      m_pFont.DrawText( 2, 40, $ffffff00, strBuffer);
    end;

    // Output text
    if m_bDisplayHelp then
    begin
      m_pFont.DrawText( 2, 80, $ffffffff, '<F1>'#10#10 +
                                          '<Up>'#10 +
                                          '<Down>'#10#10 +
                                          '<PgUp>'#10 +
                                          '<PgDn>'#10#10 +
                                          '<Home>'#10 +
                                          '<End>'#10 +
                                          '<o>', 0);
      m_pFont.DrawText(70, 80, $ffffffff, 'Toggle help'#10#10 +
                                          'Add one vertex'#10 +
                                          'Subtract one vertex'#10#10 +
                                          'Add 100 vertices'#10 +
                                          'Subtract 100 vertices'#10#10 +
                                          'Max vertices'#10 +
                                          'Min vertices'#10 +
                                          'Show optimized pmeshes');
    end else
    begin
      m_pFont.DrawText( 2, 80, $ffffffff, '<F1>');
      m_pFont.DrawText(70, 80, $ffffffff, 'Toggle help');
    end;

    // End the scene rendering
    m_pd3dDevice.EndScene;
  end;
end;

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InitDeviceObjects: HResult;
type
  TD3DXMaterialsArray = array [0..0] of TD3DXMaterial;
var
  cVerticesPerMesh: DWORD;
  pAdjacencyBuffer: ID3DXBuffer;
  pVertexBuffer:    IDirect3DVertexBuffer9;
  pMesh:            ID3DXMesh;
  pPMesh:           ID3DXPMesh;

  pTempMesh:        ID3DXMesh;
  pD3DXMtrlBuffer:  ID3DXBuffer;
  pVertices:        Pointer;
  strMediaPath:     array[0..511] of Char;
  dw32BitFlag:      DWORD;
  cVerticesMin:     DWORD;
  cVerticesMax:     DWORD;

  iPMesh: DWORD;
  Epsilons: TD3DXWeldEpsilons;
  i: Integer;
  d3dxMaterials:   ^TD3DXMaterialsArray;
  tszFilename: PChar;
{$IFDEF UNICODE}
  tszBuf: array[0..MAX_PATH-1] of Char;
{$ENDIF}
label
  _End;
begin
  // Initialize the font
  m_pFont.InitDeviceObjects(m_pd3dDevice);

  CheckMenuItem(GetMenu(m_hWnd), ID_OPTIONS_SHOWOPTIMIZEDPMESHES, IsChecked[m_bShowOptimized]);

  // Load mesh

  // Find the path to the mesh
  if FAILED(DXUtil_FindMediaFileCb(strMediaPath, SizeOf(strMediaPath), m_strMeshFilename)) then
  begin
    Result:= D3DAPPERR_MEDIANOTFOUND;
    Exit;
  end;

  // Load the mesh from the specified file
  Result:= D3DXLoadMeshFromX(strMediaPath, D3DXMESH_MANAGED, m_pd3dDevice,
                             @pAdjacencyBuffer, @pD3DXMtrlBuffer, nil,
                             @m_dwNumMaterials, pMesh);
  if FAILED(Result) then
  begin
    // hide error so that device changes will not cause exit, shows blank screen instead
    Result:= S_OK;
    goto _End;
  end;

  dw32BitFlag:= (pMesh.GetOptions and D3DXMESH_32BIT);

  // perform simple cleansing operations on mesh
  Result:= D3DXCleanMesh(pMesh, pAdjacencyBuffer.GetBufferPointer, pTempMesh,
                         pAdjacencyBuffer.GetBufferPointer, nil);
  if FAILED(Result) then
  begin
    m_dwNumMaterials := 0;
    goto _End;
  end;
  SAFE_RELEASE(pMesh);
  pMesh := pTempMesh;

  //  Perform a weld to try and remove excess vertices like the model bigship1.x in the DX8.0 SDK (current model is fixed)
  //    Weld the mesh using all epsilons of 0.0f.  A small epsilon like 1e-6 works well too
  FillChar(Epsilons, SizeOf(TD3DXWeldEpsilons), 0);
  Result:= D3DXWeldVertices(pMesh, 0, @Epsilons,
                            pAdjacencyBuffer.GetBufferPointer,
                            pAdjacencyBuffer.GetBufferPointer, nil, nil);
  if FAILED(Result) then
  begin
    m_dwNumMaterials := 0;
    goto _End;
  end;

  // verify validity of mesh for simplification
  Result:= D3DXValidMesh(pMesh, pAdjacencyBuffer.GetBufferPointer, nil);
  if FAILED(Result) then
  begin
    m_dwNumMaterials := 0;
    goto _End;
  end;

  // Allocate a material/texture arrays
  d3dxMaterials:= pD3DXMtrlBuffer.GetBufferPointer;
  GetMem(m_mtrlMeshMaterials,  m_dwNumMaterials*SizeOf(TD3DMaterial9));
  GetMem(m_pMeshTextures,      m_dwNumMaterials*SizeOf(IDirect3DTexture9));
  ZeroMemory(m_pMeshTextures,  m_dwNumMaterials*SizeOf(IDirect3DTexture9));

  // Copy the materials and load the textures
  for i:= 0 to m_dwNumMaterials - 1 do
  begin
    m_mtrlMeshMaterials[i]:= d3dxMaterials[i].MatD3D;
    m_mtrlMeshMaterials[i].Ambient:= m_mtrlMeshMaterials[i].Diffuse;

    // Find the path to the texture and create that texture
{$IFDEF UNICODE}
    tszFilename := tszBuf;
    MultiByteToWideChar(CP_ACP, 0, d3dxMaterials[i].pTextureFilename, -1, @tszBuf, MAX_PATH);
    tszBuf[MAX_PATH - 1] := #0;
{$ELSE}
    tszFilename := d3dxMaterials[i].pTextureFilename;
{$ENDIF}
    DXUtil_FindMediaFileCb(strMediaPath, SizeOf(strMediaPath), tszFilename);
    if FAILED(D3DXCreateTextureFromFile(m_pd3dDevice, strMediaPath,
                                        m_pMeshTextures[i])) then
    begin
      m_pMeshTextures[i]:= nil;
      Exit;
    end;
  end;
  pD3DXMtrlBuffer:= nil;

  // Lock the vertex buffer, to generate a simple bounding sphere
  Result:= pMesh.GetVertexBuffer(pVertexBuffer);
  if FAILED(Result) then goto _End;

  Result:= pVertexBuffer.Lock(0, 0, pVertices, D3DLOCK_NOSYSLOCK);
  if FAILED(Result) then goto _End;

  Result:= D3DXComputeBoundingSphere(PD3DXVector3(pVertices), pMesh.GetNumVertices,
                                     D3DXGetFVFVertexSize(pMesh.GetFVF),
                                     m_vObjectCenter, m_fObjectRadius);
  pVertexBuffer.Unlock;
  pVertexBuffer:= nil;

  if FAILED(Result) or (m_dwNumMaterials = 0) then
    goto _End;

  if (pMesh.GetFVF and D3DFVF_NORMAL) = 0 then
  begin
    Result:= pMesh.CloneMeshFVF(dw32BitFlag or D3DXMESH_MANAGED,
                                pMesh.GetFVF or D3DFVF_NORMAL,
                                m_pd3dDevice, pTempMesh);
    if FAILED(Result) then goto _End;

    D3DXComputeNormals(pTempMesh, nil);

    pMesh := pTempMesh;
  end;

  Result:= D3DXGeneratePMesh(pMesh, pAdjacencyBuffer.GetBufferPointer,
                             nil, nil, 1, D3DXMESHSIMP_VERTEX, pPMesh);
  if FAILED(Result) then goto _End;

  cVerticesMin := pPMesh.GetMinVertices;
  cVerticesMax := pPMesh.GetMaxVertices;

  cVerticesPerMesh := (cVerticesMax - cVerticesMin) div 10;

  m_cPMeshes := Max(1, Round(Ceil((cVerticesMax - cVerticesMin) / cVerticesPerMesh)));
  try
    GetMem(m_pPMeshes, SizeOf(ID3DXPMesh) * m_cPMeshes);
  except
    Result:= E_OUTOFMEMORY;
  end;
  if not Assigned(m_pPMeshes) then goto _End;
  FillChar(m_pPMeshes^, SizeOf(ID3DXPMesh) * m_cPMeshes, 0);

  // clone full size pmesh
  Result:= pPMesh.ClonePMeshFVF(D3DXMESH_MANAGED or D3DXMESH_VB_SHARE,
                                pPMesh.GetFVF, m_pd3dDevice, m_pPMeshFull);
  if FAILED(Result) then goto _End;

  // clone all the separate pmeshes
  for iPMesh := 0 to m_cPMeshes - 1 do
  begin
    Result:= pPMesh.ClonePMeshFVF(D3DXMESH_MANAGED or D3DXMESH_VB_SHARE,
                                  pPMesh.GetFVF, m_pd3dDevice, m_pPMeshes[iPMesh]);
    if FAILED(Result) then goto _End;

    // trim to appropriate space
    Result:= m_pPMeshes[iPMesh].TrimByVertices(cVerticesMin + cVerticesPerMesh * iPMesh,
                                               cVerticesMin + cVerticesPerMesh * (iPMesh+1), nil, nil);
    if FAILED(Result) then goto _End;

    Result:= m_pPMeshes[iPMesh].OptimizeBaseLOD(D3DXMESHOPT_VERTEXCACHE, nil);
    if FAILED(Result) then goto _End;
  end;

  // set current to be maximum number of vertices
  m_iPMeshCur := m_cPMeshes - 1;
  Result:= m_pPMeshes[m_iPMeshCur].SetNumVertices(cVerticesMax);
  if FAILED(Result) then goto _End;

  Result:= m_pPMeshFull.SetNumVertices(cVerticesMax);
  if FAILED(Result) then goto _End;

_End:
  SAFE_RELEASE(pAdjacencyBuffer);
  SAFE_RELEASE(pD3DXMtrlBuffer);
  SAFE_RELEASE(pMesh);
  SAFE_RELEASE(pPMesh);

  if FAILED(Result) then
  begin
    for iPMesh:= 0 to m_cPMeshes-1 do
      SAFE_RELEASE(m_pPMeshes[iPMesh]);

    FreeMem(m_pPMeshes);
    m_cPMeshes := 0;
    m_pPMeshes := nil;
    SAFE_RELEASE(m_pPMeshFull);
  end;
end;

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.RestoreDeviceObjects: HResult;
var
  vecLightDirUnnormalized: TD3DXVector3;
  light: TD3DLight9;
  fAspect: Single;
begin
  m_pFont.RestoreDeviceObjects;

  // Setup render state
  m_pd3dDevice.SetRenderState(D3DRS_LIGHTING,     iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_DITHERENABLE, iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_ZENABLE,      iTrue);
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);

  // Setup the light
  vecLightDirUnnormalized := D3DXVector3(0.0, -0.5, 1.0);
  ZeroMemory(@light, SizeOf(TD3DLight9));
  light._Type       := D3DLIGHT_DIRECTIONAL;
  light.Diffuse.r   := 1.0;
  light.Diffuse.g   := 1.0;
  light.Diffuse.b   := 1.0;
  D3DXVec3Normalize(light.Direction, vecLightDirUnnormalized);
  light.Position.x   := 0.0;
  light.Position.y   := -0.5;
  light.Position.z   := 1.0;
  light.Range        := 1000.0;
  m_pd3dDevice.SetLight(0, light);
  m_pd3dDevice.LightEnable(0, True);
  m_pd3dDevice.SetRenderState(D3DRS_AMBIENT, $00333333);

  m_Camera.SetWindow(m_d3dsdBackBuffer.Width, m_d3dsdBackBuffer.Height, 0.85);

  // Setup the projection matrix
  fAspect := m_d3dsdBackBuffer.Width / m_d3dsdBackBuffer.Height;
  m_Camera.SetProjParams(D3DX_PI/4, fAspect, m_fObjectRadius/64.0, m_fObjectRadius*200.0);
  m_pd3dDevice.SetTransform(D3DTS_PROJECTION, m_Camera.GetProjMatrix^);

  m_bInitDone := True;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects
// Desc:
//-----------------------------------------------------------------------------
function CMyD3DApplication.InvalidateDeviceObjects: HResult;
begin
  m_pFont.InvalidateDeviceObjects;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.DeleteDeviceObjects: HResult;
var
  i: LongWord;
  iPMesh: DWORD;
begin
  m_pFont.DeleteDeviceObjects;

  if (m_pMeshTextures <> nil) then
    for i:= 0 to m_dwNumMaterials - 1 do m_pMeshTextures[i]:= nil;

  SAFE_RELEASE(m_pPMeshFull);
  for iPMesh:= 0 to m_cPMeshes - 1 do
    SAFE_RELEASE(m_pPMeshes[iPMesh]);
  m_cPMeshes:= 0;
  FreeMem(m_pPMeshes);

  if Assigned(m_mtrlMeshMaterials) then FreeMem(m_mtrlMeshMaterials);
  m_dwNumMaterials:= 0;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FinalCleanup
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FinalCleanup: HResult;
begin
  FreeAndNil(m_pFont);
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: MsgProc
// Desc: Message proc function to handle key and menu input
//-----------------------------------------------------------------------------
function CMyD3DApplication.MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM;
  lParam: LPARAM ): LRESULT;
var
  g_strFilename: array [0..511] of Char;
  pLastSlash: PChar;
  dwNumMeshVertices: DWord;
  i: DWord;
{$WRITEABLECONST ON}
const
  ofn: TOpenFilename = (
    lStructSize: SizeOf(TOpenFilename);
    hWndOwner: 0;
    hInstance: 0;
    lpstrFilter: '.X Files (.x)'#0'*.x'#0#0;
    lpstrCustomFilter: nil;
    nMaxCustFilter: 0;
    nFilterIndex: 1;
    lpstrFile: nil; // updated by code later
    nMaxFile: 512;
    lpstrFileTitle: nil; // updated by code later
    nMaxFileTitle: 512;
    lpstrInitialDir: nil; // updated by code later
    lpstrTitle: 'Open Mesh File';
    Flags: OFN_FILEMUSTEXIST;
    nFileOffset: 0;
    nFileExtension: 1;
    lpstrDefExt: nil;
    lCustData: 0;
    lpfnHook: nil;
    lpTemplateName: nil
  );
{$WRITEABLECONST OFF}
begin
  // Pass mouse messages to the ArcBall so it can build internal matrices
  m_Camera.HandleMessages(hWnd, uMsg, wParam, lParam);

  Result:= 0;

  // Trap the context menu
  if (uMsg = WM_CONTEXTMENU) then Exit;

  if (uMsg = WM_COMMAND) then
  begin
    // Handle the about command (which displays help)
    if (LongRec(wParam).Lo = IDM_TOGGLEHELP) then
    begin
      m_bDisplayHelp:= not m_bDisplayHelp;
      Exit;
    end else if (LOWORD(wParam) = ID_OPTIONS_SHOWOPTIMIZEDPMESHES) then
    begin
      m_bShowOptimized := not m_bShowOptimized;
      CheckMenuItem(GetMenu(hWnd), ID_OPTIONS_SHOWOPTIMIZEDPMESHES, IsChecked[m_bShowOptimized]);
    end;

    // Handle the open file command
    if (m_bWindowed and (LongRec(wParam).Lo = IDM_OPENFILE)) then
    begin
      // Display the OpenFileName dialog. Then, try to load the specified file
      g_strFilename:= '';

      {$IFDEF COMPILER6_UP}
      // Handle Delphi6 case of extended OFN structure
      // (new members are invalid for Win98SE & below OS'es)
      ofn.lStructSize:= 76;

      {$ENDIF}
      ofn.lpstrFile:= m_strMeshFilename;
      ofn.lpstrFileTitle:= g_strFilename;
      ofn.lpstrInitialDir:= m_strInitialDir;

      if GetOpenFileName(ofn) then
      begin
        StrCopy(m_strInitialDir, m_strMeshFilename);
        pLastSlash:= StrPos(m_strInitialDir, '\');
        if Assigned(pLastSlash) then pLastSlash[0]:= #0;
        SetCurrentDirectory(m_strInitialDir);

        // Destroy and recreate everything
        InvalidateDeviceObjects;
        DeleteDeviceObjects;

        if FAILED(InitDeviceObjects) then
        begin
          MessageBox(m_hWnd, 'Error loading mesh: mesh may not'#10 +
                             'be valid. See debug output for'#10 +
                             'more information.'#10#10'Please select' +
                             'a different .x file.',
                             'ProgressiveMesh', MB_ICONERROR or MB_OK);
        end;

        RestoreDeviceObjects;
      end;
    end;
  end;

  if (uMsg = WM_KEYDOWN) then
  begin
    if Assigned(m_pPMeshes) then
    begin
      dwNumMeshVertices:= m_pPMeshes[m_iPMeshCur].GetNumVertices;

      if (wParam = VK_UP) then
      begin
        // Sometimes it is necessary to add more than one
        // vertex when increasing the resolution of a
        // progressive mesh, so keep adding until the
        // vertex count increases.
        for i:= 1 to 7 do
        begin
          SetNumVertices(dwNumMeshVertices+i);
          if (m_pPMeshes[m_iPMeshCur].GetNumVertices = dwNumMeshVertices + i) then Break;
        end
      end
      else if (wParam = VK_DOWN) then
          SetNumVertices(dwNumMeshVertices-1)
      else if (wParam = VK_PRIOR) then
          SetNumVertices(dwNumMeshVertices+100)
      else if (wParam = VK_NEXT) then
      begin
        if dwNumMeshVertices <= 100
          then SetNumVertices(1)
          else SetNumVertices(dwNumMeshVertices-100);
      end
      else if (wParam = VK_HOME) then
          SetNumVertices($ffffffff)
      else if (wParam = VK_END) then
          SetNumVertices(1);
    end;
  end;

  Result:= inherited MsgProc(hWnd, uMsg, wParam, lParam);
end;

//-----------------------------------------------------------------------------
// Name: SetNumVertices()
// Desc: Sets the number of vertices to display on the current progressive mesh
//-----------------------------------------------------------------------------
procedure CMyD3DApplication.SetNumVertices(dwNumVertices: DWORD);
begin
  m_pPMeshFull.SetNumVertices(dwNumVertices);

  // if current pm valid for desired value, then set the number of vertices directly
  if ((dwNumVertices >= m_pPMeshes[m_iPMeshCur].GetMinVertices) and
      (dwNumVertices <= m_pPMeshes[m_iPMeshCur].GetMaxVertices)) then
  begin
    m_pPMeshes[m_iPMeshCur].SetNumVertices( dwNumVertices);
  end else  // search for the right one
  begin
    m_iPMeshCur := m_cPMeshes - 1;

    // look for the correct "bin"
    while (m_iPMeshCur > 0) do
    begin
      // if number of vertices is less than current max then we found one to fit
      if (dwNumVertices >= m_pPMeshes[m_iPMeshCur].GetMinVertices)
      then Break;

      m_iPMeshCur:= m_iPMeshCur - 1;
    end;

    // set the vertices on the newly selected mesh
    m_pPMeshes[m_iPMeshCur].SetNumVertices(dwNumVertices);
  end;
end;

end.

