(*----------------------------------------------------------------------------*
 *  Direct3D sample from DirectX 9.0 SDK Summer 2003                          *
 *  Delphi adaptation by Alexey Barkovoy (e-mail: clootie@ixbt.com)           *
 *                                                                            *
 *  Delphi versions 5-7 are supported                                         *
 *                                                                            *
 *  Modified: 30-Dec-2003                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru/delphi                                         *
 *----------------------------------------------------------------------------*)
//-----------------------------------------------------------------------------
// File: SkinnedMesh.cpp
//
// Desc: Sample showing skinned meshes in D3D
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------

unit SkinnedMeshUnit;

interface

uses
  Windows, Messages, SysUtils,
  Direct3D9, D3DX9, D3DApp, D3DEnumeration, D3DSettings, D3DFile, D3DFont,
  D3DUtil, DXUtil;

{.$DEFINE DEBUG_VS}   // Uncomment this line to debug vertex shaders
{.$DEFINE DEBUG_PS}   // Uncomment this line to debug pixel shaders

type
  // enum for various skinning modes possible
  TMethod = (
    D3DNONINDEXED,
    D3DINDEXED,
    SOFTWARE,
    D3DINDEXEDVS,
    D3DINDEXEDHLSLVS,
    NONE
  );

  //-----------------------------------------------------------------------------
  // Name: struct D3DXFRAME_DERIVED
  // Desc: Structure derived from D3DXFRAME so we can add some app-specific
  //       info that will be stored with each frame
  //-----------------------------------------------------------------------------
  PD3DXFrameDerived = ^TD3DXFrameDerived;
  TD3DXFrameDerived = packed record { public TD3DXFrame - waiting for Delphi8 }
    Name:               PAnsiChar;
    TransformationMatrix: TD3DXMatrix;

    pMeshContainer:     PD3DXMeshContainer;

    pFrameSibling:      PD3DXFrame;
    pFrameFirstChild:   PD3DXFrame;
    ////////////////////////////////////////////
    CombinedTransformationMatrix: TD3DXMatrixA16;
  end;

  PD3DXMaterialArray = ^TD3DXMaterialArray;
  TD3DXMaterialArray = array[0..MaxInt div SizeOf(TD3DXMaterial) - 1] of TD3DXMaterial;

  PIDirect3DTexture9Array = ^TIDirect3DTexture9Array;
  TIDirect3DTexture9Array = array[0..MaxInt div SizeOf(IDirect3DTexture9) - 1] of IDirect3DTexture9;

  PD3DXMatrixPointerArray = ^TD3DXMatrixPointerArray;
  TD3DXMatrixPointerArray = array[0..MaxInt div SizeOf(Pointer) - 1] of PD3DXMatrix;

  PD3DXMatrixArray = ^TD3DXMatrixArray;
  TD3DXMatrixArray = array[0..MaxInt div SizeOf(TD3DXMatrix) - 1] of TD3DXMatrix;

  PD3DXAttributeRangeArray = ^TD3DXAttributeRangeArray;
  TD3DXAttributeRangeArray = array[0..MaxInt div SizeOf(TD3DXAttributeRange) - 1] of TD3DXAttributeRange;

  //-----------------------------------------------------------------------------
  // Name: struct D3DXMESHCONTAINER_DERIVED
  // Desc: Structure derived from D3DXMESHCONTAINER so we can add some app-specific
  //       info that will be stored with each mesh
  //-----------------------------------------------------------------------------
  PD3DXMeshContainerDerived = ^TD3DXMeshContainerDerived;
  TD3DXMeshContainerDerived = packed record { public D3DXMESHCONTAINER - waiting from Delphi8 }
    Name:               PAnsiChar;

    MeshData:           TD3DXMeshData;

    pMaterials:         PD3DXMaterialArray;
    pEffects:           PD3DXEffectInstance;
    NumMaterials:       DWORD;
    pAdjacency:         PDWORD;

    pSkinInfo:          ID3DXSkinInfo;

    pNextMeshContainer: PD3DXMeshContainer;
    ////////////////////////////////////////////
    ppTextures: PIDirect3DTexture9Array; // array of textures, entries are NULL if no texture specified

    // SkinMesh info
    pOrigMesh: ID3DXMesh;
    pAttributeTable: PD3DXAttributeRangeArray;
    NumAttributeGroups: DWORD;
    NumInfl: DWORD;
    pBoneCombinationBuf: ID3DXBuffer;
    ppBoneMatrixPtrs: PD3DXMatrixPointerArray;
    pBoneOffsetMatrices: PD3DXMatrixArray;
    NumPaletteEntries: DWORD;
    UseSoftwareVP: Boolean;
    iAttributeSW: DWORD;     // used to denote the split between SW and HW if necessary for non-indexed skinning
  end;

  CMyD3DApplication = class;

  //-----------------------------------------------------------------------------
  // Name: class CAllocateHierarchy
  // Desc: Custom version of ID3DXAllocateHierarchy with custom methods to create
  //       frames and meshcontainers.
  //-----------------------------------------------------------------------------
  CAllocateHierarchy = class(ID3DXAllocateHierarchy)
  public
    function CreateFrame(Name: PAnsiChar; out ppNewFrame: PD3DXFrame): HResult; override;
    function CreateMeshContainer(Name: PAnsiChar; const pMeshData: TD3DXMeshData;
        pMaterials: PD3DXMaterial; pEffectInstances: PD3DXEffectInstance;
        NumMaterials: DWORD; pAdjacency: PDWORD; pSkinInfo: ID3DXSkinInfo;
        out ppNewMeshContainer: PD3DXMeshContainer): HResult; override;
    function DestroyFrame(pFrameToFree: PD3DXFrame): HResult; override;
    function DestroyMeshContainer(pMeshContainerToFree: PD3DXMeshContainer): HResult; override;
    constructor Create(pApp: CMyD3DApplication); // m_pApp(pApp) {}
  public
    m_pApp: CMyD3DApplication;
  end;

  //-----------------------------------------------------------------------------
  // Name: class CMyD3DApplication
  // Desc: Application class. The base class (CD3DApplication) provides the
  //       generic functionality needed in all Direct3D samples. CMyD3DApplication
  //       adds functionality specific to this sample program.
  //-----------------------------------------------------------------------------
  CMyD3DApplication = class(CD3DApplication)
  private
    m_strMeshFilename: array[0..MAX_PATH-1] of Char;
    m_strInitialDir: array[0..MAX_PATH-1] of Char;

    m_pFrameRoot: PD3DXFrame;
    m_pAnimController: ID3DXAnimationController;
    m_pFont: CD3DFont;
    m_pFontSmall: CD3DFont;
    m_ArcBall: CD3DArcBall;          // mouse rotation utility
    m_vObjectCenter: TD3DXVector3;   // Center of bounding sphere of object
    m_fObjectRadius: Single;         // Radius of bounding sphere of object

    m_SkinningMethod: TMethod;       // Current skinning method
    m_pBoneMatrices: PD3DXMatrixArray;
    m_NumBoneMatricesMax: LongWord;

    m_pIndexedVertexShader: array[0..3] of IDirect3DVertexShader9;
    m_pEffect: ID3DXEffect;
    m_matView: TD3DXMatrixA16;
  protected
    function ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
      adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT; override;
    function OneTimeSceneInit: HResult; override;
    function InitDeviceObjects: HResult; override;
    function RestoreDeviceObjects: HResult; override;
    function InvalidateDeviceObjects: HResult; override;
    function DeleteDeviceObjects: HResult; override;
    function Render: HResult; overload; override;
    function FrameMove: HResult; override;
    function FinalCleanup: HResult; override;

    procedure DrawMeshContainer(pMeshContainerBase: PD3DXMeshContainer; pFrameBase: PD3DXFrame);
    procedure DrawFrame(pFrame: PD3DXFrame);

    function SetupBoneMatrixPointersOnMesh(pMeshContainerBase: PD3DXMeshContainer): HResult;
    function SetupBoneMatrixPointers(pFrame: PD3DXFrame): HResult;
    procedure UpdateFrameMatrices(pFrameBase: PD3DXFrame; pParentMatrix: PD3DXMatrix);
    procedure UpdateSkinningMethod(pFrameBase: PD3DXFrame);

  public
    constructor Create;
    function MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM; lParam: LPARAM ): LRESULT; override;
    function GenerateSkinnedMesh(pMeshContainer: PD3DXMeshContainerDerived): HResult;
  end;

implementation

uses
  Resource, Math, CommDlg;

const
  UINT_MAX = $FFFFFFFF; 
    
//-----------------------------------------------------------------------------
// Name: AllocateName()
// Desc: Allocates memory for a string to hold the name of a frame or mesh
//-----------------------------------------------------------------------------
//Translator: AllocateName == StrNew in Delphi
(*
HRESULT AllocateName( LPCTSTR Name, LPTSTR *pNewName )
{
    UINT cbLength;

    if (Name != NULL)
    {
        cbLength = lstrlen(Name) + 1;
        *pNewName = new TCHAR[cbLength];
        if (*pNewName == NULL)
            return E_OUTOFMEMORY;
        memcpy(*pNewName, Name, cbLength*sizeof(TCHAR));
    }
    else
    {
        *pNewName = NULL;
    }

    return S_OK;
}
*)

{ CAllocateHierarchy }

constructor CAllocateHierarchy.Create(pApp: CMyD3DApplication);
begin
  m_pApp:= pApp;
end;

//-----------------------------------------------------------------------------
// Name: CAllocateHierarchy::CreateFrame()
// Desc:
//-----------------------------------------------------------------------------
function CAllocateHierarchy.CreateFrame(Name: PAnsiChar;
  out ppNewFrame: PD3DXFrame): HResult;
var
  pFrame: PD3DXFrameDerived;
begin
  Result := S_OK;

  ppNewFrame := nil;
  New(pFrame); // {PD3DXFrameDerived}
  try try
    pFrame.Name:= StrNew(Name);

    // initialize other data members of the frame
    D3DXMatrixIdentity(pFrame.TransformationMatrix);
    D3DXMatrixIdentity(pFrame.CombinedTransformationMatrix);

    pFrame.pMeshContainer := nil;
    pFrame.pFrameSibling := nil;
    pFrame.pFrameFirstChild := nil;

    ppNewFrame := PD3DXFrame(pFrame);
    pFrame := nil;
  except
    on EOutOfMemory do Result:= E_OUTOFMEMORY;
    else raise;
  end;
  finally
    Dispose(pFrame);
  end;
end;

//-----------------------------------------------------------------------------
// Name: CAllocateHierarchy::CreateMeshContainer()
// Desc:
//-----------------------------------------------------------------------------
function CAllocateHierarchy.CreateMeshContainer(Name: PAnsiChar;
  const pMeshData: TD3DXMeshData; pMaterials: PD3DXMaterial;
  pEffectInstances: PD3DXEffectInstance; NumMaterials: DWORD;
  pAdjacency: PDWORD; pSkinInfo: ID3DXSkinInfo;
  out ppNewMeshContainer: PD3DXMeshContainer): HResult;
var
  pMeshContainer: PD3DXMeshContainerDerived;
  NumFaces: LongWord;
  iMaterial: Integer;
  iBone, cBones: Integer;
  pd3dDevice: IDirect3DDevice9;

  pMesh: ID3DXMesh;
  strTexturePath: array[0..MAX_PATH-1] of Char;
  tszFilename: PChar;
{$IFDEF UNICODE}
  tszBuf: array[0..MAX_PATH-1] of Char;
{$ENDIF}
begin
  pMeshContainer := nil;
//  pd3dDevice := nil; pMesh := nil; // Translator - not needed in Delphi

  ppNewMeshContainer := nil;

  Result:= E_FAIL;

  try try
    // this sample does not handle patch meshes, so fail when one is found
    if (pMeshData._Type <> D3DXMESHTYPE_MESH) then Exit;

    // get the pMesh interface pointer out of the mesh data structure
    pMesh := pMeshData.pMesh as ID3DXMesh;

    // this sample does not FVF compatible meshes, so fail when one is found
    if (pMesh.GetFVF = 0) then Exit;

    // allocate the overloaded structure to return as a D3DXMESHCONTAINER

    New(pMeshContainer); // = new D3DXMESHCONTAINER_DERIVED;
    FillChar(pMeshContainer^, 0, SizeOf(TD3DXMeshContainerDerived));

    // make sure and copy the name.  All memory as input belongs to caller, interfaces can be addref'd though
    pMeshContainer.Name := StrNew(Name);

    pMesh.GetDevice(pd3dDevice);
    NumFaces := pMesh.GetNumFaces;

    // if no normals are in the mesh, add them
    if ((pMesh.GetFVF and D3DFVF_NORMAL) = 0) then
    begin
      pMeshContainer.MeshData._Type := D3DXMESHTYPE_MESH;

      // clone the mesh to make room for the normals
      Result := pMesh.CloneMeshFVF(pMesh.GetOptions,
                                   pMesh.GetFVF or D3DFVF_NORMAL,
                                   pd3dDevice, ID3DXMesh(pMeshContainer.MeshData.pMesh));
      if FAILED(Result) then Exit;

      // get the new pMesh pointer back out of the mesh container to use
      // NOTE: we do not release pMesh because we do not have a reference to it yet
      pMesh := pMeshContainer.MeshData.pMesh as ID3DXMesh;;

      // now generate the normals for the pmesh
      D3DXComputeNormals(pMesh, nil);
    end else  // if no normals, just add a reference to the mesh for the mesh container
    begin
      pMeshContainer.MeshData.pMesh := pMesh;
      pMeshContainer.MeshData._Type := D3DXMESHTYPE_MESH;
    end;

    // allocate memory to contain the material information.  This sample uses
    //   the D3D9 materials and texture names instead of the EffectInstance style materials
    pMeshContainer.NumMaterials := Max(1, NumMaterials);

    GetMem(pMeshContainer.pMaterials, SizeOf(TD3DXMaterial)*pMeshContainer.NumMaterials);
    GetMem(pMeshContainer.ppTextures, SizeOf(IDirect3DTexture9)*pMeshContainer.NumMaterials);
    GetMem(pMeshContainer.pAdjacency, SizeOf(DWORD)*NumFaces*3);

    Move(pAdjacency^, pMeshContainer.pAdjacency^, SizeOf(DWORD)*NumFaces*3);
    FillChar(pMeshContainer.ppTextures^, SizeOf(IDirect3DTexture9)*pMeshContainer.NumMaterials, 0);

    // if materials provided, copy them
    if (NumMaterials > 0) then
    begin
      Move(pMaterials^, pMeshContainer.pMaterials^, SizeOf(TD3DXMaterial)*NumMaterials);

      for iMaterial := 0 to NumMaterials - 1 do
      begin
        if (pMeshContainer.pMaterials[iMaterial].pTextureFilename <> nil) then
        begin
          strTexturePath:= '';
{$IFDEF UNICODE}
          tszFilename := tszBuf;
          MultiByteToWideChar(CP_ACP, 0, pMeshContainer.pMaterials[iMaterial].pTextureFilename, -1, @tszBuf, MAX_PATH);
          tszBuf[MAX_PATH - 1] := #0;
{$ELSE}
          tszFilename := pMeshContainer.pMaterials[iMaterial].pTextureFilename;
{$ENDIF}
          DXUtil_FindMediaFileCb(strTexturePath, SizeOf(strTexturePath), tszFilename);
          if FAILED(D3DXCreateTextureFromFile(pd3dDevice, strTexturePath,
                                              pMeshContainer.ppTextures[iMaterial]))
          then pMeshContainer.ppTextures[iMaterial] := nil;


          // don't remember a pointer into the dynamic memory, just forget the name after loading
          pMeshContainer.pMaterials[iMaterial].pTextureFilename := nil;
        end;
      end;
    end else // if no materials provided, use a default one
    begin
        pMeshContainer.pMaterials[0].pTextureFilename := nil;
        FillChar(pMeshContainer.pMaterials[0].MatD3D, SizeOf(TD3DMaterial9), 0);
        pMeshContainer.pMaterials[0].MatD3D.Diffuse.r := 0.5;
        pMeshContainer.pMaterials[0].MatD3D.Diffuse.g := 0.5;
        pMeshContainer.pMaterials[0].MatD3D.Diffuse.b := 0.5;
        pMeshContainer.pMaterials[0].MatD3D.Specular := pMeshContainer.pMaterials[0].MatD3D.Diffuse;
    end;

    // if there is skinning information, save off the required data and then setup for HW skinning
    if (pSkinInfo <> nil) then
    begin
      // first save off the SkinInfo and original mesh data
      pMeshContainer.pSkinInfo := pSkinInfo;
      pMeshContainer.pOrigMesh := pMesh;

      // Will need an array of offset matrices to move the vertices from the figure space to the bone's space
      cBones := pSkinInfo.GetNumBones;
      GetMem(pMeshContainer.pBoneOffsetMatrices, SizeOf(TD3DXMatrix)*cBones);

      // get each of the bone offset matrices so that we don't need to get them later
      for iBone := 0 to cBones - 1 do
      begin
        pMeshContainer.pBoneOffsetMatrices[iBone] := pMeshContainer.pSkinInfo.GetBoneOffsetMatrix(iBone)^;
      end;

      // GenerateSkinnedMesh will take the general skinning information and transform it to a HW friendly version
      Result := m_pApp.GenerateSkinnedMesh(pMeshContainer);
      if FAILED(Result) then Exit;
    end;

    ppNewMeshContainer := PD3DXMeshContainer(pMeshContainer);
    pMeshContainer := nil;
    Result:= S_OK;
  except
    on EOutOfMemory do Result:= E_OUTOFMEMORY;
    else raise;
  end;
  finally
    SAFE_RELEASE(pd3dDevice);

    // call Destroy function to properly clean up the memory allocated
    if (pMeshContainer <> nil) then DestroyMeshContainer(PD3DXMeshContainer(pMeshContainer));
  end;
end;

//-----------------------------------------------------------------------------
// Name: CAllocateHierarchy::DestroyFrame()
// Desc:
//-----------------------------------------------------------------------------
function CAllocateHierarchy.DestroyFrame(
  pFrameToFree: PD3DXFrame): HResult;
begin
  StrDispose(pFrameToFree.Name);
  Dispose(pFrameToFree);
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: CAllocateHierarchy::DestroyMeshContainer()
// Desc:
//-----------------------------------------------------------------------------
function CAllocateHierarchy.DestroyMeshContainer(
  pMeshContainerToFree: PD3DXMeshContainer): HResult;
var
  iMaterial: Integer;
  pMeshContainer: PD3DXMeshContainerDerived;
begin
  pMeshContainer := PD3DXMeshContainerDerived(pMeshContainerToFree);

  StrDispose(pMeshContainer.Name);
  FreeMem(pMeshContainer.pAdjacency);
  FreeMem(pMeshContainer.pMaterials);
  FreeMem(pMeshContainer.pBoneOffsetMatrices);

  // release all the allocated textures
  if (pMeshContainer.ppTextures <> nil) then
  begin
    for iMaterial := 0 to pMeshContainer.NumMaterials - 1 do
    begin
      pMeshContainer.ppTextures[iMaterial]:= nil;
    end;
  end;

  FreeMem(pMeshContainer.ppTextures);
  FreeMem(pMeshContainer.ppBoneMatrixPtrs);
  pMeshContainer.pBoneCombinationBuf:= nil;
  pMeshContainer.MeshData.pMesh:= nil;
  pMeshContainer.pSkinInfo:= nil;
  pMeshContainer.pOrigMesh:= nil;
  Dispose(pMeshContainer);
  Result:= S_OK;
end;


{ CMyD3DApplication }

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
constructor CMyD3DApplication.Create;
begin
  inherited Create;

  m_strWindowTitle    := 'Skinned Mesh';
  m_d3dEnumeration.AppUsesDepthBuffer   := True;
  m_d3dEnumeration.AppUsesMixedVP := True;
  m_bShowCursorWhenFullscreen := True;

  m_pFont         := CD3DFont.Create('Arial', 12, D3DFONT_BOLD);
  m_pFontSmall    := CD3DFont.Create('Arial',  9, D3DFONT_BOLD);

  m_pAnimController := nil;
  m_pFrameRoot := nil;

  DXUtil_GetDXSDKMediaPathCb(m_strInitialDir, SizeOf(m_strInitialDir));
  StrCopy(m_strMeshFilename, 'tiny.x');

  m_SkinningMethod := D3DNONINDEXED;
  m_pBoneMatrices := nil;
  m_NumBoneMatricesMax := 0;

  m_ArcBall:= CD3DArcBall.Create;
end;

//-----------------------------------------------------------------------------
// Name: GenerateSkinnedMesh()
// Desc: Called either by CreateMeshContainer when loading a skin mesh, or when
//       changing methods.  This function uses the pSkinInfo of the mesh
//       container to generate the desired drawable mesh and bone combination
//       table.
//-----------------------------------------------------------------------------
function CMyD3DApplication.GenerateSkinnedMesh(
  pMeshContainer: PD3DXMeshContainerDerived): HResult;
var
  rgBoneCombinations: PD3DXBoneCombination;
  cInfl: DWORD;
  i: Integer;
  iInfl: Integer;
  pMeshTmp: ID3DXMesh;

  NumMaxFaceInfl: DWORD;
  Flags: DWORD;
  pIB: IDirect3DIndexBuffer9;
  MaxMatrices: LongWord;
  NewFVF: DWORD;
  pMesh: ID3DXMesh;
  pDecl: TFVFDeclaration; // array[0..MAX_FVF_DECL_SIZE-1] of TD3DVertexElement9;
  pDeclCur: PD3DVertexElement9;
begin
  Result:= S_OK;

  if (pMeshContainer.pSkinInfo = nil) then Exit;

  SAFE_RELEASE(pMeshContainer.MeshData.pMesh);
  SAFE_RELEASE(pMeshContainer.pBoneCombinationBuf);

  // if non-indexed skinning mode selected, use ConvertToBlendedMesh to generate drawable mesh
  if (m_SkinningMethod = D3DNONINDEXED) then
  begin
    Result := pMeshContainer.pSkinInfo.ConvertToBlendedMesh(
                                 pMeshContainer.pOrigMesh,
                                 D3DXMESH_MANAGED or D3DXMESHOPT_VERTEXCACHE,
                                 pMeshContainer.pAdjacency,
                                 nil, nil, nil,
                                 @pMeshContainer.NumInfl,
                                 pMeshContainer.NumAttributeGroups,
                                 pMeshContainer.pBoneCombinationBuf,
                                 ID3DXMesh(pMeshContainer.MeshData.pMesh));
    if FAILED(Result) then Exit;

    (* If the device can only do 2 matrix blends, ConvertToBlendedMesh cannot approximate all meshes to it
       Thus we split the mesh in two parts: The part that uses at most 2 matrices and the rest. The first is
       drawn using the device's HW vertex processing and the rest is drawn using SW vertex processing. *)
    rgBoneCombinations := PD3DXBoneCombination(pMeshContainer.pBoneCombinationBuf.GetBufferPointer);

    // look for any set of bone combinations that do not fit the caps
    for i := 0 to pMeshContainer.NumAttributeGroups - 1 do
    begin
      pMeshContainer.iAttributeSW := i;
      cInfl := 0;

      for iInfl := 0 to pMeshContainer.NumInfl - 1 do
      begin
        if (rgBoneCombinations.BoneId[iInfl] <> UINT_MAX)
        then Inc(cInfl);
      end;
      Inc(rgBoneCombinations);

      if (cInfl > m_d3dCaps.MaxVertexBlendMatrices) then Break;
    end;

    // if there is both HW and SW, add the Software Processing flag
    if (pMeshContainer.iAttributeSW < pMeshContainer.NumAttributeGroups) then
    begin
      with ID3DXMesh(pMeshContainer.MeshData.pMesh) do
      Result := CloneMeshFVF(
                    D3DXMESH_SOFTWAREPROCESSING or GetOptions,
                    GetFVF,
                    m_pd3dDevice, pMeshTmp);
      if FAILED(Result) then Exit;

      pMeshContainer.MeshData.pMesh:= nil;
      pMeshContainer.MeshData.pMesh := pMeshTmp;
      pMeshTmp := nil;
    end;
  end
  // if indexed skinning mode selected, use ConvertToIndexedsBlendedMesh to generate drawable mesh
  else if (m_SkinningMethod = D3DINDEXED) then
  begin
    Flags := D3DXMESHOPT_VERTEXCACHE;

    Result := pMeshContainer.pOrigMesh.GetIndexBuffer(pIB);
    if FAILED(Result) then Exit;

    Result := pMeshContainer.pSkinInfo.GetMaxFaceInfluences(pIB, pMeshContainer.pOrigMesh.GetNumFaces, NumMaxFaceInfl);
    pIB := nil;
    if FAILED(Result) then Exit;

    // 12 entry palette guarantees that any triangle (4 independent influences per vertex of a tri)
    // can be handled
    NumMaxFaceInfl := Min(NumMaxFaceInfl, 12);

    if (m_d3dCaps.MaxVertexBlendMatrixIndex + 1 < NumMaxFaceInfl) then
    begin
      // HW does not support indexed vertex blending. Use SW instead
      pMeshContainer.NumPaletteEntries := min(256, pMeshContainer.pSkinInfo.GetNumBones);
      pMeshContainer.UseSoftwareVP := True;
      Flags := Flags or D3DXMESH_SYSTEMMEM;
    end else
    begin
      // using hardware - determine palette size from caps and number of bones
      // If normals are present in the vertex data that needs to be blended for lighting, then
      // the number of matrices is half the number specified by MaxVertexBlendMatrixIndex.
      pMeshContainer.NumPaletteEntries := Min( (m_d3dCaps.MaxVertexBlendMatrixIndex + 1) div 2,
                                               pMeshContainer.pSkinInfo.GetNumBones);
      pMeshContainer.UseSoftwareVP := False;
      Flags := Flags or D3DXMESH_MANAGED;
    end;

    Result := pMeshContainer.pSkinInfo.ConvertToIndexedBlendedMesh(
                                            pMeshContainer.pOrigMesh,
                                            Flags,
                                            pMeshContainer.NumPaletteEntries,
                                            pMeshContainer.pAdjacency,
                                            nil, nil, nil,
                                            @pMeshContainer.NumInfl,
                                            pMeshContainer.NumAttributeGroups,
                                            pMeshContainer.pBoneCombinationBuf,
                                            ID3DXMesh(pMeshContainer.MeshData.pMesh));
    if FAILED(Result) then Exit;
  end
  // if vertex shader indexed skinning mode selected, use ConvertToIndexedsBlendedMesh to generate drawable mesh
  else if ((m_SkinningMethod = D3DINDEXEDVS) or (m_SkinningMethod = D3DINDEXEDHLSLVS)) then
  begin
    // Get palette size
    // First 9 constants are used for other data.  Each 4x3 matrix takes up 3 constants.
    // (96 - 9) /3 i.e. Maximum constant count - used constants
    MaxMatrices := 26;
    pMeshContainer.NumPaletteEntries := Min(MaxMatrices, pMeshContainer.pSkinInfo.GetNumBones);

    Flags := D3DXMESHOPT_VERTEXCACHE;
    if (m_d3dCaps.VertexShaderVersion >= D3DVS_VERSION(1, 1)) then
    begin
      pMeshContainer.UseSoftwareVP := False;
      Flags := Flags or D3DXMESH_MANAGED;
    end else
    begin
      pMeshContainer.UseSoftwareVP := True;
      Flags := Flags or D3DXMESH_SYSTEMMEM;
    end;

    SAFE_RELEASE(pMeshContainer.MeshData.pMesh);

    Result := pMeshContainer.pSkinInfo.ConvertToIndexedBlendedMesh(
                                            pMeshContainer.pOrigMesh,
                                            Flags,
                                            pMeshContainer.NumPaletteEntries,
                                            pMeshContainer.pAdjacency,
                                            nil, nil, nil,
                                            @pMeshContainer.NumInfl,
                                            pMeshContainer.NumAttributeGroups,
                                            pMeshContainer.pBoneCombinationBuf,
                                            ID3DXMesh(pMeshContainer.MeshData.pMesh));
    if FAILED(Result) then Exit;


    // FVF has to match our declarator. Vertex shaders are not as forgiving as FF pipeline
    NewFVF := (ID3DXMesh(pMeshContainer.MeshData.pMesh).GetFVF and D3DFVF_POSITION_MASK) or D3DFVF_NORMAL or D3DFVF_TEX1 or D3DFVF_LASTBETA_UBYTE4;
    if (NewFVF <> ID3DXMesh(pMeshContainer.MeshData.pMesh).GetFVF) then
    begin
      Result := ID3DXMesh(pMeshContainer.MeshData.pMesh).CloneMeshFVF(
                    ID3DXMesh(pMeshContainer.MeshData.pMesh).GetOptions, NewFVF, m_pd3dDevice, pMesh);
      if FAILED(Result) then
      begin
        // pMeshContainer.MeshData.pMesh := nil
        pMeshContainer.MeshData.pMesh := pMesh;
        pMesh := nil;
      end;
    end;

    Result := ID3DXMesh(pMeshContainer.MeshData.pMesh).GetDeclaration(pDecl);
    if FAILED(Result) then Exit;

    // the vertex shader is expecting to interpret the UBYTE4 as a D3DCOLOR, so update the type
    //   NOTE: this cannot be done with CloneMesh, that would convert the UBYTE4 data to float and then to D3DCOLOR
    //          this is more of a "cast" operation
    pDeclCur := @pDecl;
    while (pDeclCur.Stream <> $ff) do
    begin
      if ((pDeclCur.Usage = D3DDECLUSAGE_BLENDINDICES) and (pDeclCur.UsageIndex = 0))
        then pDeclCur._Type := D3DDECLTYPE_D3DCOLOR;
      Inc(pDeclCur);
    end;

    Result := ID3DXMesh(pMeshContainer.MeshData.pMesh).UpdateSemantics(pDecl);
    if FAILED(Result) then Exit;

    // allocate a buffer for bone matrices, but only if another mesh has not allocated one of the same size or larger
    if (m_NumBoneMatricesMax < pMeshContainer.pSkinInfo.GetNumBones) then
    begin
      m_NumBoneMatricesMax := pMeshContainer.pSkinInfo.GetNumBones;

      // Allocate space for blend matrices
      Dispose(m_pBoneMatrices);
      GetMem(m_pBoneMatrices, SizeOf(TD3DXMatrixA16)*m_NumBoneMatricesMax);
    end;

  end
  // if software skinning selected, use GenerateSkinnedMesh to create a mesh that can be used with UpdateSkinnedMesh
  else if (m_SkinningMethod = SOFTWARE) then
  begin
    Result := pMeshContainer.pOrigMesh.CloneMeshFVF(D3DXMESH_MANAGED, pMeshContainer.pOrigMesh.GetFVF,
                                          m_pd3dDevice, ID3DXMesh(pMeshContainer.MeshData.pMesh));
    if FAILED(Result) then Exit;

    Result := ID3DXMesh(pMeshContainer.MeshData.pMesh).GetAttributeTable(nil, @pMeshContainer.NumAttributeGroups);
    if FAILED(Result) then Exit;

    Dispose(pMeshContainer.pAttributeTable);
    GetMem(pMeshContainer.pAttributeTable, SizeOf(TD3DXAttributeRange)*pMeshContainer.NumAttributeGroups);

    Result := ID3DXMesh(pMeshContainer.MeshData.pMesh).GetAttributeTable(@pMeshContainer.pAttributeTable[0], nil);
    if FAILED(Result) then Exit;

    // allocate a buffer for bone matrices, but only if another mesh has not allocated one of the same size or larger
    if (m_NumBoneMatricesMax < pMeshContainer.pSkinInfo.GetNumBones) then
    begin
      m_NumBoneMatricesMax := pMeshContainer.pSkinInfo.GetNumBones;

      // Allocate space for blend matrices
      Dispose(m_pBoneMatrices);
      GetMem(m_pBoneMatrices, SizeOf(TD3DXMatrixA16)*m_NumBoneMatricesMax);
    end;
  end
  else  // invalid m_SkinningMethod value
  begin
    // return failure due to invalid skinning method value
    Result := E_INVALIDARG;
  end;

end;

//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
function CMyD3DApplication.OneTimeSceneInit: HResult;
begin
    // Set cursor to indicate that user can move the object with the mouse
{$IFDEF WIN64}
  SetClassLongPtr(m_hWnd, GCLP_HCURSOR, LONG_PTR(LoadCursor(nil, IDC_SIZEALL));
{$ELSE}
  SetClassLong(m_hWnd, GCL_HCURSOR, LoadCursor(0, IDC_SIZEALL));
{$ENDIF}
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FrameMove: HResult;
var
  matWorld: TD3DXMatrixA16;
  vEye, vAt, vUp: TD3DXVector3;
begin
  // Setup world matrix
  D3DXMatrixTranslation(matWorld, -m_vObjectCenter.x,
                                  -m_vObjectCenter.y,
                                  -m_vObjectCenter.z);
  D3DXMatrixMultiply(matWorld, matWorld, m_ArcBall.GetRotationMatrix^);
  D3DXMatrixMultiply(matWorld, matWorld, m_ArcBall.GetTranslationMatrix^);
  m_pd3dDevice.SetTransform(D3DTS_WORLD, matWorld);

  vEye:= D3DXVector3(0, 0, -2*m_fObjectRadius);
  vAt:= D3DXVector3(0, 0, 0);
  vUp:= D3DXVector3(0, 1, 0);
  D3DXMatrixLookAtLH(m_matView, vEye, vAt, vUp);

  m_pd3dDevice.SetTransform(D3DTS_VIEW, m_matView);

  if (m_pAnimController <> nil) then
    m_pAnimController.AdvanceTime(m_fElapsedTime, nil);

  UpdateFrameMatrices(m_pFrameRoot, @matWorld);

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: DrawMeshContainer()
// Desc: Called to render a mesh in the hierarchy
//-----------------------------------------------------------------------------
procedure CMyD3DApplication.DrawMeshContainer(
  pMeshContainerBase: PD3DXMeshContainer; pFrameBase: PD3DXFrame);
type
  PD3DXBoneCombinationArray = ^TD3DXBoneCombinationArray;
  TD3DXBoneCombinationArray = array[0..MaxInt div SizeOf(TD3DXBoneCombination)-1] of TD3DXBoneCombination;
var
  pMeshContainer: PD3DXMeshContainerDerived;
  pFrame: PD3DXFrameDerived;
  iMaterial: LongWord;
  NumBlend: LongWord;
  iAttrib: Integer;
  AttribIdPrev: DWORD;
  pBoneComb: PD3DXBoneCombinationArray;

  iMatrixIndex: LongWord;
  iPaletteEntry: LongWord;
  matTemp: TD3DXMatrixA16;

  i: Integer;
  vConst: TD3DXVector4;

  color1, color2, ambEmm: TD3DXColor;
  numPasses: Integer;

  Identity:  TD3DXMatrix;
  cBones: DWORD;
  iBone: DWORD;
  pbVerticesSrc: PByte;
  pbVerticesDest: PByte;
  iPass: Integer;
begin
  pMeshContainer := PD3DXMeshContainerDerived(pMeshContainerBase);
  pFrame := PD3DXFrameDerived(pFrameBase);

  // first check for skinning
  if (pMeshContainer.pSkinInfo <> nil) then
  begin
    if (m_SkinningMethod = D3DNONINDEXED) then
    begin
      AttribIdPrev := UNUSED32;
      pBoneComb := PD3DXBoneCombinationArray(pMeshContainer.pBoneCombinationBuf.GetBufferPointer);

      // Draw using default vtx processing of the device (typically HW)
      for iAttrib := 0 to pMeshContainer.NumAttributeGroups - 1 do
      begin
        NumBlend := 0;
        for i := 0 to pMeshContainer.NumInfl - 1 do
        begin
          if (pBoneComb[iAttrib].BoneId[i] <> UINT_MAX) then
          begin
            NumBlend := i;
          end;
        end;

        if (m_d3dCaps.MaxVertexBlendMatrices >= NumBlend + 1) then
        begin
          // first calculate the world matrices for the current set of blend weights and get the accurate count of the number of blends
          for i := 0 to pMeshContainer.NumInfl - 1 do
          begin
            iMatrixIndex := pBoneComb[iAttrib].BoneId[i];
            if (iMatrixIndex <> UINT_MAX) then
            begin
              D3DXMatrixMultiply(matTemp, pMeshContainer.pBoneOffsetMatrices[iMatrixIndex], pMeshContainer.ppBoneMatrixPtrs[iMatrixIndex]^);
              m_pd3dDevice.SetTransform(D3DTS_WORLDMATRIX(i), matTemp);
            end;
          end;

          m_pd3dDevice.SetRenderState(D3DRS_VERTEXBLEND, NumBlend);

          // lookup the material used for this subset of faces
          if ((AttribIdPrev <> pBoneComb[iAttrib].AttribId) or (AttribIdPrev = UNUSED32)) then
          begin
            m_pd3dDevice.SetMaterial(pMeshContainer.pMaterials[pBoneComb[iAttrib].AttribId].MatD3D);
            m_pd3dDevice.SetTexture(0, pMeshContainer.ppTextures[pBoneComb[iAttrib].AttribId]);
            AttribIdPrev := pBoneComb[iAttrib].AttribId;
          end;

          // draw the subset now that the correct material and matrices are loaded
          ID3DXMesh(pMeshContainer.MeshData.pMesh).DrawSubset(iAttrib);
        end;
      end;

      // If necessary, draw parts that HW could not handle using SW
      if (pMeshContainer.iAttributeSW < pMeshContainer.NumAttributeGroups) then
      begin
        AttribIdPrev := UNUSED32;
        m_pd3dDevice.SetSoftwareVertexProcessing(TRUE);
        for iAttrib := pMeshContainer.iAttributeSW to pMeshContainer.NumAttributeGroups - 1 do
        begin
          NumBlend := 0;
          for i := 0 to pMeshContainer.NumInfl - 1 do
          begin
            if (pBoneComb[iAttrib].BoneId[i] <> UINT_MAX)
            then NumBlend := i;
          end;

          if (m_d3dCaps.MaxVertexBlendMatrices < NumBlend + 1) then
          begin
            // first calculate the world matrices for the current set of blend weights and get the accurate count of the number of blends
            for i := 0 to pMeshContainer.NumInfl - 1 do
            begin
              iMatrixIndex := pBoneComb[iAttrib].BoneId[i];
              if (iMatrixIndex <> UINT_MAX) then
              begin
                D3DXMatrixMultiply(matTemp, pMeshContainer.pBoneOffsetMatrices[iMatrixIndex], pMeshContainer.ppBoneMatrixPtrs[iMatrixIndex]^);
                m_pd3dDevice.SetTransform(D3DTS_WORLDMATRIX(i), matTemp);
              end;
            end;

            m_pd3dDevice.SetRenderState(D3DRS_VERTEXBLEND, NumBlend);

            // lookup the material used for this subset of faces
            if ((AttribIdPrev <> pBoneComb[iAttrib].AttribId) or (AttribIdPrev = UNUSED32)) then
            begin
              m_pd3dDevice.SetMaterial(pMeshContainer.pMaterials[pBoneComb[iAttrib].AttribId].MatD3D);
              m_pd3dDevice.SetTexture(0, pMeshContainer.ppTextures[pBoneComb[iAttrib].AttribId]);
              AttribIdPrev := pBoneComb[iAttrib].AttribId;
            end;

            // draw the subset now that the correct material and matrices are loaded
            ID3DXMesh(pMeshContainer.MeshData.pMesh).DrawSubset(iAttrib);
          end;
        end;
        m_pd3dDevice.SetSoftwareVertexProcessing(False);
      end;

      m_pd3dDevice.SetRenderState(D3DRS_VERTEXBLEND, 0);
    end
    else if (m_SkinningMethod = D3DINDEXED) then
    begin
      // if hw doesn't support indexed vertex processing, switch to software vertex processing
      if (pMeshContainer.UseSoftwareVP)
      then m_pd3dDevice.SetSoftwareVertexProcessing(True);

      // set the number of vertex blend indices to be blended
      if (pMeshContainer.NumInfl = 1)
      then m_pd3dDevice.SetRenderState(D3DRS_VERTEXBLEND, D3DVBF_0WEIGHTS)
      else m_pd3dDevice.SetRenderState(D3DRS_VERTEXBLEND, pMeshContainer.NumInfl - 1);

      if (pMeshContainer.NumInfl <> 0)
      then m_pd3dDevice.SetRenderState(D3DRS_INDEXEDVERTEXBLENDENABLE, iTrue);

      // for each attribute group in the mesh, calculate the set of matrices in the palette and then draw the mesh subset
      pBoneComb := PD3DXBoneCombinationArray(pMeshContainer.pBoneCombinationBuf.GetBufferPointer);
      for iAttrib := 0 to pMeshContainer.NumAttributeGroups - 1 do
      begin
        // first calculate all the world matrices
        for iPaletteEntry := 0 to pMeshContainer.NumPaletteEntries - 1 do
        begin
          iMatrixIndex := pBoneComb[iAttrib].BoneId[iPaletteEntry];
          if (iMatrixIndex <> UINT_MAX) then
          begin
            D3DXMatrixMultiply(matTemp, pMeshContainer.pBoneOffsetMatrices[iMatrixIndex], pMeshContainer.ppBoneMatrixPtrs[iMatrixIndex]^);
            m_pd3dDevice.SetTransform(D3DTS_WORLDMATRIX(iPaletteEntry), matTemp);
          end;
        end;

        // setup the material of the mesh subset - REMEMBER to use the original pre-skinning attribute id to get the correct material id
        m_pd3dDevice.SetMaterial(pMeshContainer.pMaterials[pBoneComb[iAttrib].AttribId].MatD3D);
        m_pd3dDevice.SetTexture(0, pMeshContainer.ppTextures[pBoneComb[iAttrib].AttribId]);

        // finally draw the subset with the current world matrix palette and material state
        ID3DXMesh(pMeshContainer.MeshData.pMesh).DrawSubset(iAttrib);
      end;

      // reset blending state
      m_pd3dDevice.SetRenderState(D3DRS_INDEXEDVERTEXBLENDENABLE, iFalse);
      m_pd3dDevice.SetRenderState(D3DRS_VERTEXBLEND, 0);

      // remember to reset back to hw vertex processing if software was required
      if (pMeshContainer.UseSoftwareVP)
      then m_pd3dDevice.SetSoftwareVertexProcessing(False);
    end
    else if (m_SkinningMethod = D3DINDEXEDVS) then
    begin
      // Use COLOR instead of UBYTE4 since Geforce3 does not support it
      // vConst.w should be 3, but due to COLOR/UBYTE4 issue, mul by 255 and add epsilon
      vConst:= D3DXVector4(1.0, 0.0, 0.0, 765.01);

      if (pMeshContainer.UseSoftwareVP)
      then m_pd3dDevice.SetSoftwareVertexProcessing(True);

      m_pd3dDevice.SetVertexShader(m_pIndexedVertexShader[pMeshContainer.NumInfl - 1]);

      pBoneComb := PD3DXBoneCombinationArray(pMeshContainer.pBoneCombinationBuf.GetBufferPointer);
      for iAttrib := 0 to pMeshContainer.NumAttributeGroups - 1 do
      begin
        // first calculate all the world matrices
        for iPaletteEntry := 0 to pMeshContainer.NumPaletteEntries - 1 do
        begin
          iMatrixIndex := pBoneComb[iAttrib].BoneId[iPaletteEntry];
          if (iMatrixIndex <> UINT_MAX) then
          begin
            D3DXMatrixMultiply(matTemp, pMeshContainer.pBoneOffsetMatrices[iMatrixIndex], pMeshContainer.ppBoneMatrixPtrs[iMatrixIndex]^);
            D3DXMatrixMultiplyTranspose(matTemp, matTemp, m_matView);
            m_pd3dDevice.SetVertexShaderConstantF(iPaletteEntry*3 + 9, PSingle(@matTemp), 3);
          end;
        end;

        // Sum of all ambient and emissive contribution
        color1:= pMeshContainer.pMaterials[pBoneComb[iAttrib].AttribId].MatD3D.Ambient;
        color2:= D3DXColor(0.25, 0.25, 0.25, 1.0);
        D3DXColorModulate(ambEmm, color1, color2);
        D3DXColorAdd(ambEmm, ambEmm, pMeshContainer.pMaterials[pBoneComb[iAttrib].AttribId].MatD3D.Emissive);

        // set material color properties
        m_pd3dDevice.SetVertexShaderConstantF(8, PSingle(@pMeshContainer.pMaterials[pBoneComb[iAttrib].AttribId].MatD3D.Diffuse), 1);
        m_pd3dDevice.SetVertexShaderConstantF(7, PSingle(@ambEmm), 1);
        vConst.y := pMeshContainer.pMaterials[pBoneComb[iAttrib].AttribId].MatD3D.Power;
        m_pd3dDevice.SetVertexShaderConstantF(0, PSingle(@vConst), 1);

        m_pd3dDevice.SetTexture(0, pMeshContainer.ppTextures[pBoneComb[iAttrib].AttribId]);

        // finally draw the subset with the current world matrix palette and material state
        ID3DXMesh(pMeshContainer.MeshData.pMesh).DrawSubset(iAttrib);

      end;

      // remember to reset back to hw vertex processing if software was required
      if (pMeshContainer.UseSoftwareVP)
        then m_pd3dDevice.SetSoftwareVertexProcessing(False);
      m_pd3dDevice.SetVertexShader(nil);
    end
    else if (m_SkinningMethod = D3DINDEXEDHLSLVS) then
    begin
      if (pMeshContainer.UseSoftwareVP)
      then m_pd3dDevice.SetSoftwareVertexProcessing(True);

      pBoneComb := PD3DXBoneCombinationArray(pMeshContainer.pBoneCombinationBuf.GetBufferPointer);
      for iAttrib := 0 to pMeshContainer.NumAttributeGroups - 1 do
      begin
        // first calculate all the world matrices
        for iPaletteEntry := 0 to pMeshContainer.NumPaletteEntries - 1 do
        begin
          iMatrixIndex := pBoneComb[iAttrib].BoneId[iPaletteEntry];
          if (iMatrixIndex <> UINT_MAX) then
          begin
            D3DXMatrixMultiply(matTemp, pMeshContainer.pBoneOffsetMatrices[iMatrixIndex], pMeshContainer.ppBoneMatrixPtrs[iMatrixIndex]^);
            D3DXMatrixMultiply(m_pBoneMatrices[iPaletteEntry], matTemp, m_matView);
          end;
        end;
        m_pEffect.SetMatrixArray('mWorldMatrixArray', @m_pBoneMatrices[0], pMeshContainer.NumPaletteEntries);

        // Sum of all ambient and emissive contribution
        color1:= (pMeshContainer.pMaterials[pBoneComb[iAttrib].AttribId].MatD3D.Ambient);
        color2:= D3DXColor(0.25, 0.25, 0.25, 1.0);
        D3DXColorModulate(ambEmm, color1, color2);
        D3DXColorAdd(ambEmm, ambEmm, pMeshContainer.pMaterials[pBoneComb[iAttrib].AttribId].MatD3D.Emissive);

        // set material color properties
        m_pEffect.SetVector('MaterialDiffuse', PD3DXVector4(@pMeshContainer.pMaterials[pBoneComb[iAttrib].AttribId].MatD3D.Diffuse)^);
        m_pEffect.SetVector('MaterialAmbient', PD3DXVector4(@ambEmm)^);

        // setup the material of the mesh subset - REMEMBER to use the original pre-skinning attribute id to get the correct material id
        m_pd3dDevice.SetTexture(0, pMeshContainer.ppTextures[pBoneComb[iAttrib].AttribId]);

        // Set CurNumBones to select the correct vertex shader for the number of bones
        m_pEffect.SetInt('CurNumBones', pMeshContainer.NumInfl - 1);

        // Start the effect now all parameters have been updated
        m_pEffect._Begin(@numPasses, D3DXFX_DONOTSAVESTATE);
        for iPass := 0 to numPasses - 1 do
        begin
          m_pEffect.Pass(iPass);
          // draw the subset with the current world matrix palette and material state
          ID3DXMesh(pMeshContainer.MeshData.pMesh).DrawSubset(iAttrib);
        end;

        m_pEffect._End;

        m_pd3dDevice.SetVertexShader(nil);
      end;

      // remember to reset back to hw vertex processing if software was required
      if (pMeshContainer.UseSoftwareVP)
      then m_pd3dDevice.SetSoftwareVertexProcessing(False);
    end
    else if (m_SkinningMethod = SOFTWARE) then
    begin
      cBones := pMeshContainer.pSkinInfo.GetNumBones;

      // set up bone transforms
      for iBone := 0 to cBones - 1 do
      begin
        D3DXMatrixMultiply
        (
           m_pBoneMatrices[iBone],                 // output
           pMeshContainer.pBoneOffsetMatrices[iBone],
           pMeshContainer.ppBoneMatrixPtrs[iBone]^
        );
      end;

      // set world transform
      D3DXMatrixIdentity(Identity);
      m_pd3dDevice.SetTransform(D3DTS_WORLD, Identity);

      pMeshContainer.pOrigMesh.LockVertexBuffer(D3DLOCK_READONLY, Pointer(pbVerticesSrc));
      ID3DXMesh(pMeshContainer.MeshData.pMesh).LockVertexBuffer(0, Pointer(pbVerticesDest));

      // generate skinned mesh
      pMeshContainer.pSkinInfo.UpdateSkinnedMesh(@m_pBoneMatrices[0], nil, pbVerticesSrc, pbVerticesDest);

      pMeshContainer.pOrigMesh.UnlockVertexBuffer;
      ID3DXMesh(pMeshContainer.MeshData.pMesh).UnlockVertexBuffer;

      for iAttrib := 0 to pMeshContainer.NumAttributeGroups - 1 do
      begin
        m_pd3dDevice.SetMaterial(pMeshContainer.pMaterials[pMeshContainer.pAttributeTable[iAttrib].AttribId].MatD3D);
        m_pd3dDevice.SetTexture(0, pMeshContainer.ppTextures[pMeshContainer.pAttributeTable[iAttrib].AttribId]);
        ID3DXMesh(pMeshContainer.MeshData.pMesh).DrawSubset(pMeshContainer.pAttributeTable[iAttrib].AttribId);
      end;
    end
    else // bug out as unsupported mode
    begin
      Exit;
    end;
  end
  else  // standard mesh, just draw it after setting material properties
  begin
    m_pd3dDevice.SetTransform(D3DTS_WORLD, pFrame.CombinedTransformationMatrix);

    for iMaterial := 0 to pMeshContainer.NumMaterials - 1 do
    begin
      m_pd3dDevice.SetMaterial(pMeshContainer.pMaterials[iMaterial].MatD3D);
      m_pd3dDevice.SetTexture(0, pMeshContainer.ppTextures[iMaterial]);
      ID3DXMesh(pMeshContainer.MeshData.pMesh).DrawSubset(iMaterial);
    end;
  end;
end;


//-----------------------------------------------------------------------------
// Name: DrawFrame()
// Desc: Called to render a frame in the hierarchy
//-----------------------------------------------------------------------------
procedure CMyD3DApplication.DrawFrame(pFrame: PD3DXFrame);
var
  pMeshContainer: PD3DXMeshContainer;
begin
  pMeshContainer := pFrame.pMeshContainer;
  while (pMeshContainer <> nil) do
  begin
    DrawMeshContainer(pMeshContainer, pFrame);

    pMeshContainer := pMeshContainer.pNextMeshContainer;
  end;

  if (pFrame.pFrameSibling <> nil)
    then DrawFrame(pFrame.pFrameSibling);

  if (pFrame.pFrameFirstChild <> nil)
    then DrawFrame(pFrame.pFrameFirstChild);
end;


//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.Render: HResult;
var
  vecLightDirUnnormalized: TD3DXVector3;
  light: TD3DLight9;
  vLightDir: TD3DXVector4;
  matProj: TD3DXMatrixA16;
  fAspect: Single;
  pstrType: PChar;
begin
  // Clear the backbuffer
  m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER, $000000ff, 1.0, 0);

  // Setup the light
  vecLightDirUnnormalized := D3DXVector3(0.0, -1.0, 1.0);
  ZeroMemory(@light, SizeOf(TD3DLight9));
  light._Type       := D3DLIGHT_DIRECTIONAL;
  light.Diffuse.r   := 1.0;
  light.Diffuse.g   := 1.0;
  light.Diffuse.b   := 1.0;
  D3DXVec3Normalize(light.Direction, vecLightDirUnnormalized);
  light.Position.x   := 0.0;
  light.Position.y   := -1.0;
  light.Position.z   := 1.0;
  light.Range        := 1000.0;
  m_pd3dDevice.SetLight(0, light);
  m_pd3dDevice.LightEnable(0, True);

  // Set Light for vertex shader
  vLightDir:= D3DXVECTOR4(0.0, 1.0, -1.0, 0.0);
  D3DXVec4Normalize(vLightDir, vLightDir);
  m_pd3dDevice.SetVertexShaderConstantF(1, PSingle(@vLightDir), 1);

  // Setup the projection matrix
  fAspect := m_d3dsdBackBuffer.Width / m_d3dsdBackBuffer.Height;
  D3DXMatrixPerspectiveFovLH(matProj, D3DX_PI/4, fAspect,
                             m_fObjectRadius/64.0, m_fObjectRadius*200.0);
  m_pd3dDevice.SetTransform(D3DTS_PROJECTION, matProj);

  m_pEffect.SetMatrix('mViewProj', matProj);
  m_pEffect.SetVector('lhtDir', vLightDir);

  // set the projection matrix for the vertex shader based skinning method
  D3DXMatrixTranspose(matProj, matProj);
  m_pd3dDevice.SetVertexShaderConstantF(2, PSingle(@matProj), 4);

  // Begin the scene
  if SUCCEEDED(m_pd3dDevice.BeginScene) then
  begin
    DrawFrame(m_pFrameRoot);

    // Output statistics
    m_pFont.DrawText( 2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats);
    m_pFont.DrawText( 2, 20, D3DCOLOR_ARGB(255,255,255,0), m_strDeviceStats);
    if (m_SkinningMethod = D3DNONINDEXED) then
        pstrType := 'Using non-indexed skinning'
    else if( m_SkinningMethod = D3DINDEXED) then
      pstrType := 'Using indexed skinning'
    else if( m_SkinningMethod = SOFTWARE) then
      pstrType := 'Using software skinning'
    else if( m_SkinningMethod = D3DINDEXEDVS) then
      pstrType := 'Using indexed skinning with a vertex shader'
    else if( m_SkinningMethod = D3DINDEXEDHLSLVS) then
      pstrType := 'Using indexed skinning with a HLSL vertex shader'
    else
      pstrType := 'No skinning';
    m_pFontSmall.DrawText( 2, 40, D3DCOLOR_ARGB(255,255,255,255), pstrType);

    // End the scene.
    m_pd3dDevice.EndScene;
  end;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: SetupBoneMatrixPointersOnMesh()
// Desc: Called to setup the pointers for a given bone to its transformation matrix
//-----------------------------------------------------------------------------
function CMyD3DApplication.SetupBoneMatrixPointersOnMesh(
  pMeshContainerBase: PD3DXMeshContainer): HResult;
var
  iBone, cBones: LongWord;
  pFrame: PD3DXFrameDerived;
  pMeshContainer: PD3DXMeshContainerDerived;
begin
  pMeshContainer := PD3DXMeshContainerDerived(pMeshContainerBase);

  // if there is a skinmesh, then setup the bone matrices
  if (pMeshContainer.pSkinInfo <> nil) then
  begin
    cBones := pMeshContainer.pSkinInfo.GetNumBones;

    GetMem(pMeshContainer.ppBoneMatrixPtrs, SizeOf(TD3DXMatrix)*cBones);

    for iBone := 0 to cBones - 1 do
    begin
      pFrame := PD3DXFrameDerived(D3DXFrameFind(m_pFrameRoot, pMeshContainer.pSkinInfo.GetBoneName(iBone)));
      if (pFrame = nil) then
      begin
        Result:= E_FAIL;
        Exit;
      end;

      pMeshContainer.ppBoneMatrixPtrs[iBone] := @pFrame.CombinedTransformationMatrix;
    end;
  end;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: SetupBoneMatrixPointers()
// Desc: Called to setup the pointers for a given bone to its transformation matrix
//-----------------------------------------------------------------------------
function CMyD3DApplication.SetupBoneMatrixPointers(
  pFrame: PD3DXFrame): HResult;
begin
  if (pFrame.pMeshContainer <> nil) then
  begin
    Result := SetupBoneMatrixPointersOnMesh(pFrame.pMeshContainer);
    if FAILED(Result) then Exit;
  end;

  if (pFrame.pFrameSibling <> nil) then
  begin
    Result := SetupBoneMatrixPointers(pFrame.pFrameSibling);
    if FAILED(Result) then Exit;
  end;

  if (pFrame.pFrameFirstChild <> nil) then
  begin
    Result := SetupBoneMatrixPointers(pFrame.pFrameFirstChild);
    if FAILED(Result) then Exit;
  end;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: UpdateFrameMatrices()
// Desc: update the frame matrices
//-----------------------------------------------------------------------------
procedure CMyD3DApplication.UpdateFrameMatrices(pFrameBase: PD3DXFrame; pParentMatrix: PD3DXMatrix);
var
  pFrame: PD3DXFrameDerived;
begin
  pFrame := PD3DXFrameDerived(pFrameBase);

  if (pParentMatrix <> nil) then
    D3DXMatrixMultiply(pFrame.CombinedTransformationMatrix, pFrame.TransformationMatrix, pParentMatrix^)
  else
    pFrame.CombinedTransformationMatrix := pFrame.TransformationMatrix;

  if (pFrame.pFrameSibling <> nil) then
    UpdateFrameMatrices(pFrame.pFrameSibling, pParentMatrix);

  if (pFrame.pFrameFirstChild <> nil) then
    UpdateFrameMatrices(pFrame.pFrameFirstChild, @pFrame.CombinedTransformationMatrix);
end;


//-----------------------------------------------------------------------------
// Name: UpdateSkinningMethod()
// Desc: update the skinning method
//-----------------------------------------------------------------------------
procedure CMyD3DApplication.UpdateSkinningMethod(pFrameBase: PD3DXFrame);
var
  pFrame: PD3DXFrameDerived;
  pMeshContainer: PD3DXMeshContainerDerived;
begin
  pFrame := PD3DXFrameDerived(pFrameBase);

  pMeshContainer := PD3DXMeshContainerDerived(pFrame.pMeshContainer);

  while (pMeshContainer <> nil) do
  begin
    GenerateSkinnedMesh(pMeshContainer);

    pMeshContainer := PD3DXMeshContainerDerived(pMeshContainer.pNextMeshContainer);
  end;

  if (pFrame.pFrameSibling <> nil) then
    UpdateSkinningMethod(pFrame.pFrameSibling);

  if (pFrame.pFrameFirstChild <> nil) then
    UpdateSkinningMethod(pFrame.pFrameFirstChild);
end;


//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: This creates all device-dependent managed objects, such as managed
//       textures and managed vertex buffers.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InitDeviceObjects: HResult;
var
  strMeshPath: array[0..MAX_PATH-1] of Char;
  strSkinnedMeshFXPath: array[0..MAX_PATH-1] of Char;
  Alloc: CAllocateHierarchy;
  dwShaderFlags: DWORD;
begin
  Alloc:= CAllocateHierarchy.Create(Self);

  // Initialize the font's internal textures
  m_pFont.InitDeviceObjects(m_pd3dDevice);
  m_pFontSmall.InitDeviceObjects(m_pd3dDevice);

  // Load the mesh from the specified file
  Result := DXUtil_FindMediaFileCb(strMeshPath, SizeOf(strMeshPath), m_strMeshFilename);
  if FAILED(Result) then Exit;

  Result := D3DXLoadMeshHierarchyFromX(strMeshPath, D3DXMESH_MANAGED, m_pd3dDevice, Alloc, nil, m_pFrameRoot, m_pAnimController);
  if FAILED(Result) then Exit;

  Result := SetupBoneMatrixPointers(m_pFrameRoot);
  if FAILED(Result) then Exit;

  Result := D3DXFrameCalculateBoundingSphere(m_pFrameRoot, m_vObjectCenter, m_fObjectRadius);
  if FAILED(Result) then Exit;

  // Define DEBUG_VS and/or DEBUG_PS to debug vertex and/or pixel shaders with the shader debugger.
  // Debugging vertex shaders requires either REF or software vertex processing, and debugging
  // pixel shaders requires REF.  The D3DXSHADER_FORCE_*_SOFTWARE_NOOPT flag improves the debug
  // experience in the shader debugger.  It enables source level debugging, prevents instruction
  // reordering, prevents dead code elimination, and forces the compiler to compile against the next
  // higher available software target, which ensures that the unoptimized shaders do not exceed
  // the shader model limitations.  Setting these flags will cause slower rendering since the shaders
  // will be unoptimized and forced into software.  See the DirectX documentation for more information
  // about using the shader debugger.
  dwShaderFlags := 0;
  {$IFDEF DEBUG_VS}
     dwShaderFlags := dwShaderFlags or D3DXSHADER_FORCE_VS_SOFTWARE_NOOPT;
  {$ENDIF}
  {$IFDEF DEBUG_PS}
     dwShaderFlags := dwShaderFlags or D3DXSHADER_FORCE_PS_SOFTWARE_NOOPT;
  {$ENDIF}

  // Create Effect for HLSL skinning
  // Find the vertex shader file
  Result := DXUtil_FindMediaFileCb(strSkinnedMeshFXPath, SizeOf(strSkinnedMeshFXPath), 'SkinnedMesh.fx');
  if FAILED(Result) then Exit;

  Result := D3DXCreateEffectFromFile(m_pd3dDevice, strSkinnedMeshFXPath, nil,
                                     nil, dwShaderFlags, nil, m_pEffect, nil);
  if FAILED(Result) then Exit;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Restore device-memory objects and state after a device is created or
//       resized.
//-----------------------------------------------------------------------------
function CMyD3DApplication.RestoreDeviceObjects: HResult;
var
  iInfl: DWORD;
  pCode: ID3DXBuffer;
begin
  // Setup render state
  m_pd3dDevice.SetRenderState(D3DRS_LIGHTING,         iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_DITHERENABLE,     iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_ZENABLE,          iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_CULLMODE,         D3DCULL_CCW);
  m_pd3dDevice.SetRenderState(D3DRS_AMBIENT,          $33333333);
  m_pd3dDevice.SetRenderState(D3DRS_NORMALIZENORMALS, iTrue);
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);

  // Setup the arcball parameters
  m_ArcBall.SetWindow(m_d3dsdBackBuffer.Width, m_d3dsdBackBuffer.Height, 0.85);
  m_ArcBall.SetTranslationRadius(m_fObjectRadius);

  // Restore Effect
  if (m_pEffect <> nil) then m_pEffect.OnResetDevice;

  // Restore the font
  m_pFont.RestoreDeviceObjects;
  m_pFontSmall.RestoreDeviceObjects;

  // load the indexed vertex shaders
  for iInfl := 0 to 3 do
  begin
    // Assemble the vertex shader file
    Result := D3DXAssembleShaderFromResource(0, MAKEINTRESOURCE(IDD_SHADER1 + iInfl), nil, nil, 0, @pCode, nil);
    if FAILED(Result) then Exit;

    Result := D3DXAssembleShaderFromResource(0, MAKEINTRESOURCE(IDD_SHADER1 + iInfl), nil, nil, 0, @pCode, nil);
    if FAILED(Result) then Exit;

    // Create the vertex shader
    Result := m_pd3dDevice.CreateVertexShader(PDWORD(pCode.GetBufferPointer),
                                              m_pIndexedVertexShader[iInfl]);
    if FAILED(Result) then Exit;

    pCode:= nil;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc: Called when the device-dependent objects are about to be lost.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InvalidateDeviceObjects: HResult;
var
  iInfl: DWORD;
begin
  m_pFont.InvalidateDeviceObjects;
  m_pFontSmall.InvalidateDeviceObjects;

  // release the vertex shaders
  for iInfl := 0 to 3 do
    SAFE_RELEASE(m_pIndexedVertexShader[iInfl]);

  // Invalidate Effect
  if (m_pEffect <> nil) then m_pEffect.OnLostDevice;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.DeleteDeviceObjects: HResult;
var
  Alloc: CAllocateHierarchy;
begin
  Alloc:= CAllocateHierarchy.Create(Self);

  m_pFont.DeleteDeviceObjects;
  m_pFontSmall.DeleteDeviceObjects;

  D3DXFrameDestroy(m_pFrameRoot, Alloc);

  SAFE_RELEASE(m_pAnimController);
  SAFE_RELEASE(m_pEffect);

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FinalCleanup: HResult;
begin
  SAFE_DELETE(m_pFont);
  SAFE_DELETE(m_pFontSmall);
  m_ArcBall.Free;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Message proc function to handle key and menu input
//-----------------------------------------------------------------------------
function CMyD3DApplication.MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM;
  lParam: LPARAM): LRESULT;
var
  ofn: TOpenFilename;
  g_strFilename: array[0..MAX_PATH-1] of Char;
  pLastSlash: PChar;
  hMenu: THandle; // HMENU;
  NewSkinningMethod: TMethod;
begin
  // Pass mouse messages to the ArcBall so it can build internal matrices
  m_ArcBall.HandleMessages(hWnd, uMsg, wParam, lParam);

  // Trap context menu
  if (WM_CONTEXTMENU = uMsg) then
  begin
    Result:= 0;
    Exit;
  end;

  // Handle key presses
  if (WM_KEYDOWN = uMsg) then
  begin
    if ('W' = Char(wParam)) then
      m_pd3dDevice.SetRenderState(D3DRS_FILLMODE, D3DFILL_WIREFRAME);

    if ('S' = Char(wParam)) then
      m_pd3dDevice.SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID);
  end
  else if (uMsg = WM_COMMAND) then
  begin
    // Handle the open file command
    if (LOWORD(wParam) = IDM_OPENFILE) then
    begin
      g_strFilename:= '';

      // Display the OpenFileName dialog. Then, try to load the specified file
      FillChar(ofn, SizeOf(ofn), 0);
      ofn.lStructSize       := SizeOf(ofn);
      ofn.hwndOwner         := m_hWnd;
      ofn.hInstance         := 0; //m_hInstance;
      ofn.lpstrFilter       := '.X Files (.x)'#0'*.x'#0#0;
      ofn.lpstrCustomFilter := nil;
      ofn.nMaxCustFilter    := 0;
      ofn.nFilterIndex      := 1;
      ofn.lpstrFile         := m_strMeshFilename;
      ofn.nMaxFile          := MAX_PATH;
      ofn.lpstrFileTitle    := g_strFilename;
      ofn.nMaxFileTitle     := MAX_PATH;
      ofn.lpstrInitialDir   := m_strInitialDir;
      ofn.lpstrTitle        := 'Open Mesh File';
      ofn.Flags             := OFN_FILEMUSTEXIST;
      ofn.nFileOffset       := 0;
      ofn.nFileExtension    := 1;
      // ...
{      ofn.lpstrDefExt       := nil; // '*.x';
      ofn.lCustData         := 0;
      ofn.lpfnHook          := nil;
      ofn.lpTemplateName    := nil;
      ofn.pvReserved        := nil;
      ofn.dwReserved        := 0;
      ofn.FlagsEx           := 0; }

      {$IFDEF COMPILER6_UP}
      // Handle Delphi6 & Up case of extended OFN structure
      // (new members are invalid for Win98SE & below OS'es)
      ofn.lStructSize:= 76;

      {$ENDIF}
      if GetOpenFileName(ofn) then
      begin
        StrCopy(m_strInitialDir, m_strMeshFilename);
        pLastSlash := StrRScan(m_strInitialDir, '\');
        if (pLastSlash <> nil) then pLastSlash^ := #0;
        SetCurrentDirectory(m_strInitialDir);

        // Destroy and recreate everything
        InvalidateDeviceObjects;
        DeleteDeviceObjects;
        InitDeviceObjects;
        RestoreDeviceObjects;
      end;
    end
    else if ((LOWORD(wParam) >= ID_OPTIONS_D3DNONINDEXED) and ((LOWORD(wParam) <= ID_OPTIONS_HLSLINDEXED))) then
    begin
      hMenu := GetMenu(hWnd);
      NewSkinningMethod := TMethod(LOWORD(wParam) - ID_OPTIONS_D3DNONINDEXED);

      CheckMenuItem(hMenu, ID_OPTIONS_D3DNONINDEXED,    MF_UNCHECKED);
      CheckMenuItem(hMenu, ID_OPTIONS_D3DINDEXED,       MF_UNCHECKED);
      CheckMenuItem(hMenu, ID_OPTIONS_SOFTWARESKINNING, MF_UNCHECKED);
      CheckMenuItem(hMenu, ID_OPTIONS_D3DINDEXEDVS,     MF_UNCHECKED);
      CheckMenuItem(hMenu, ID_OPTIONS_HLSLINDEXED,      MF_UNCHECKED);
      CheckMenuItem(hMenu, LOWORD(wParam), MF_CHECKED);

      // if the selected skinning method is different than the current one
      if (m_SkinningMethod <> NewSkinningMethod) then
      begin
        m_SkinningMethod := NewSkinningMethod;

        // update the meshes to the new skinning method
        UpdateSkinningMethod(m_pFrameRoot);
      end;
    end;
  end;

  // Pass remaining messages to default handler
  Result:= inherited MsgProc(hWnd, uMsg, wParam, lParam);
end;


//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device initialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
function CMyD3DApplication.ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
      adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT;
begin
  Result:= E_FAIL;

  // Always request MIXED or SOFTWARE vertex processing do deal with fall backs
  if (dwBehavior and D3DCREATE_PUREDEVICE <> 0) or
     (dwBehavior and D3DCREATE_HARDWARE_VERTEXPROCESSING <> 0) 
  then Exit;

  // If using mixed, make sure the hardware can do vertex blending
  if (dwBehavior and D3DCREATE_MIXED_VERTEXPROCESSING <> 0) and
     (pCaps.MaxVertexBlendMatrices < 2) 
  then Exit;

  Result:= S_OK;
end;

end.

